module dst_mkl_lib
  use precision   ! exports kind value r_kind
  use mkl_dfti    ! mkl dft
  
  implicit none
  
  interface dst_mkl
    module procedure dst
  end interface dst_mkl

  public :: dst_mkl, dst_mkl_finalize

  private

  ! List data types
  type ll_dst_node
     integer :: n
     type(DFTI_DESCRIPTOR), pointer ::  Desc_Handle 
     type(ll_dst_node), pointer :: next => null()
  end type ll_dst_node

  type ll_dst_list
     type(ll_dst_node) :: head
  end type ll_dst_list

  type(ll_dst_list) :: ll_init_sizes


contains
  
  subroutine dst( w ) 
    
    !     dst performs a Discrete Sine Tranformation to a vector w. 
    !
    !     This routine uses the mlk.
    !     The Transformation is NOT normalized. Two calls to this routine
    !     multiplies vector w by 2*(n+1).
    !
    !     The first call to this routine initializes some vectors and workspace
    !     used in the subsequent calls with the same size. 
    !     A call with a different size produces a new initialization
    !
    !     This module mus be compiled with openmp flag if will be use with
    !     conclurrent threads
    !
    real(kind=r_kind), dimension(:), intent(inout), target :: w

    integer :: n, ierr
    
    n = size(w,1)
    
    !$OMP CRITICAL    
    if (.not. ll_dst_exists_n(ll_init_sizes, n)) then
      ierr = init_dst(n)
      if (ierr .ne. 0) stop 'Error initializing dst()'
    end if
    !$OMP END CRITICAL    
    
    call apply_dst(w)

    return
    
  end subroutine dst
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  

  integer function init_dst(n) 
    integer, intent(in) :: n

    init_dst = ll_dst_add_n_dst(ll_init_sizes, n)

  end function init_dst

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  

  subroutine apply_dst(w) 
    real(kind=r_kind), dimension(:), intent(inout) :: w

    integer :: n, info
    real(kind=r_kind), dimension(2*(size(w,1)+1)) :: v

    type(ll_dst_node), pointer :: p_node
   
    n = size(w,1)
    p_node => ll_dst_get_n_node(ll_init_sizes, n)
    if (.not. associated(p_node)) then
       print *, 'Error: trying to work with a non-initialized problem size'
       return
    end if

    v = (/0D+1,w,0D+1,-w(n:1:-1)/)
    info = DftiComputeForward( p_node%Desc_Handle, v )
    call DftierrorManager( info )

    w = -v(3:2*(n+1):2)

  end subroutine apply_dst

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  

  integer function ll_dst_add_n_dst(ll, n)
    type(ll_dst_list), intent(inout) :: ll
    integer, intent(in) :: n

    type(ll_dst_node), pointer :: node

    integer :: info

    if (ll_dst_exists_n(ll, n)) then
       ll_dst_add_n_dst = 1
       return
    end if

    allocate(node, stat=info)
    if (info .ne. 0) then
       print *,'Error allocating memory in ll_dst_add_n_dst'
       ll_dst_add_n_dst = -1
    end if

    node%n = n

    info = DftiCreateDescriptor( node%Desc_Handle, DFTI_DOUBLE, DFTI_REAL, 1, 2*(node%n+1) )
    call DftierrorManager( info )
    !info = DftiSetValue( node%Desc_Handle, DFTI_FORWARD_SCALE, 1.0/sqrt((n+1)/2.0) )
    !call DftierrorManager( info )
    info = DftiSetValue( node%Desc_Handle, DFTI_PACKED_FORMAT, DFTI_PACK_FORMAT )
    call DftierrorManager( info )
    info = DftiCommitDescriptor( node%Desc_Handle )
    call DftierrorManager( info )

    node%next => ll%head%next
    ll%head%next => node

    ll_dst_add_n_dst = 0

  end function ll_dst_add_n_dst

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  logical function ll_dst_exists_n(ll, n)
    type(ll_dst_list), intent(in), target :: ll
    integer, intent(in) :: n

    type(ll_dst_node), pointer :: p_node
    
    ll_dst_exists_n = .false.

    p_node => ll%head
    do while (associated(p_node%next))
       p_node => p_node%next
       if (p_node%n .eq. n) then
          ll_dst_exists_n = .true.
          exit
       end if
    end do

  end function ll_dst_exists_n

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  function ll_dst_get_n_node(ll,n)
    type(ll_dst_node), pointer :: ll_dst_get_n_node
    type(ll_dst_list), intent(in), target :: ll
    integer, intent(in) :: n

    ll_dst_get_n_node => ll%head
    do while (associated(ll_dst_get_n_node%next))
       ll_dst_get_n_node => ll_dst_get_n_node%next
       if ( ll_dst_get_n_node%n .eq. n) then
          return
       end if
    end do

    nullify(ll_dst_get_n_node)

  end function ll_dst_get_n_node

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  

  subroutine DftierrorManager( Status )
    integer, intent(in) :: Status
    integer :: error
    if( DftiErrorClass( Status, error ) .eq. .true. ) then
      print *, DftiErrorMessage( Status )
      stop
    end if
  end subroutine DftierrorManager

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  

  integer function dst_mkl_finalize()

    dst_mkl_finalize =  ll_dst_finalize(ll_init_sizes)

  end function dst_mkl_finalize

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  integer function ll_dst_finalize(ll)
    type(ll_dst_list), intent(inout), target :: ll
    integer :: i, n, ret
    
    n = ll_dst_length(ll)
    
    do i=1, n
       ret = ll_dst_remove_node(ll,1)
       if (ret .ne. 0) then
          ll_dst_finalize = -1
          return
       end if
    end do
    
    ll_dst_finalize = 0
    
  end function ll_dst_finalize
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function ll_dst_length(ll)
    type(ll_dst_list), intent(in), target :: ll
    type(ll_dst_node), pointer :: p_node
    integer :: n

    p_node => ll%head
    n=0
    do while (associated(p_node%next))
       p_node => p_node%next
       n = n+1
    end do

    ll_dst_length = n

  end function ll_dst_length

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function ll_dst_remove_node(ll, i)
    type(ll_dst_list), intent(inout), target :: ll
    integer, intent(in) :: i

    type(ll_dst_node), pointer :: p_node, p_ant
    integer :: j, info
    
    p_ant  => ll%head
    p_node => ll%head%next
    do j=1,(i-1)
       p_ant => p_node
       p_node => p_node%next
    end do

    p_ant%next => p_node%next

    deallocate(p_node, stat=info)
    if (info .ne. 0) then
       print *,'Error deallocating memory in ll_dst_remove_node'
       ll_dst_remove_node = -2
    end if

    ll_dst_remove_node = 0

  end function ll_dst_remove_node
  
end module dst_mkl_lib
