!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This module applies the Gokhberg-Sementsul formula
! See Bunch 1985
! Only the first formula at the moment
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
module gokh_sement_lib
  use precision   ! exports kind value r_kind
  use rft_mkl     
  
  implicit none
  
  public :: gokh_sement, gokh_sement_init, gokh_sement_apply 
  private

contains
  
  ! This routine applies the formula to a vector v
  ! to compute the product inv(T)*v, where T is Toeplitz
  subroutine gokh_sement( x, v, y_, w_, new_matrix_, only_init_ ) 
    !
    !     x   is the column of inv(T)
    !     y_  is the row of inv(T)
    !     v   is the array to be multiplied
    !     w_  is the array solution so v won't be written
    !     new_matrix_  is for the initialization of the arrays
    !     only_init_  is for the initialization of the arrays only. Must be used with new_matrix_=.true.
    !                 In this case array v is not referenced.
    !
    !     NOTE: This routine can be used only for 
    !           one Toeplitz matrix at a time
    !
    real(kind=r_kind), dimension(:),         intent(in)            :: x
    real(kind=r_kind), dimension(:), target, intent(inout)         :: v
    real(kind=r_kind), dimension(:),         intent(in),  optional :: y_
    real(kind=r_kind), dimension(:), target, intent(out), optional :: w_
    logical,                                 intent(in),  optional :: new_matrix_, only_init_

    real(kind=r_kind), dimension(:), allocatable, save    :: rft_x, rft_y
    real(kind=r_kind), dimension(:), allocatable          :: rft_v, z1, z2
    real(kind=r_kind), dimension(:), pointer              :: w

    logical :: new_matrix, only_init
    real(kind=r_kind), save :: alpha
    real(kind=r_kind) :: eps

    integer :: n
    logical :: non_symmetric
    
    integer :: o, info, i

    n = size(x,1)
    non_symmetric = present( y_ )
    
    ! Testing the precision
    eps = epsilon(eps)
    if( eps*10000.gt.abs(x(1)) ) then
      print *,'gokh_sement warning: possible precision error (',x(1),')'
    end if

    if( size(v,1).ne.n ) then
      print *,'gokh_sement: Dimensions error'
      return
    end if

    if( non_symmetric ) then
      if( size(y_,1).ne.n ) then
        print *,'gokh_sement: Dimensions error'
        return
      end if
      print *,'gokh_sement: Non symmetric version still not implemented'
      return
    end if 
   
    if( present(w_) ) then
      w => w_
    else
      w => v
    end if
 
    if( present( new_matrix_ ) ) then
      new_matrix = new_matrix_
    else 
      new_matrix = .false.
    end if

    if( present( only_init_ ) ) then
      if( .not.new_matrix ) then
        print *,'gokh_sement: only_init_ must be used with new_matrix=.true.'
        return
      end if
      only_init = only_init_
    else 
      only_init = .false.
    end if

    ! This routine doesn't use the conv module because it is 
    ! This routine doesn't use the conv module because it is 
    ! a particular case. It stores the state from the first call to de rest

    o = 2
    do while( o.lt.(2*n-1) ) 
      o = 2 * o
    end do

    allocate( rft_v(o), z1(o), z2(o), stat=info)
    if (info .ne. 0) then
       print *,'Error allocating memory in gokh_sement'
       return
    end if      
 
    if( new_matrix ) then 
      if( allocated( rft_x ).and.allocated( rft_y ) ) then
        deallocate( rft_x, rft_y, stat=info)
        if (info .ne. 0) then
           print *,'Error deallocating memory in gokh_sement'
           return
        end if      
      end if
      allocate( rft_x(o), rft_y(o), stat=info)
      if (info .ne. 0) then
         print *,'Error allocating memory in gokh_sement'
         return
      end if      
 
      alpha = ONE / ( x(1) * o * o )

      ! rft_x is the diagonal of the lower triangular factor of first member of the GS formula
      rft_x = (/x,(ZERO,i=1,o-n)/)
      call rft( rft_x )

      ! rft_y is the diagonal of the lower triangular factor of second member of the GS formula
      rft_y = (/ZERO,(x(n:2:-1)),(ZERO,i=1,o-n)/)
      call rft( rft_y )
      if( only_init ) return
    else 
      if( .not.allocated( rft_x ) .or. .not.allocated( rft_y ) ) then
        print *,'First call to gokh_sement must be done with argument new_matrix_=.true.'
        return
      end if
    end if ! first call

    rft_v = alpha * (/v,(ZERO,i=1,o-n)/) 
    call rft( rft_v )
    call rft_product( rft_y, rft_v, z2, .true. )
    call irft( z2 )
    z2(n+1:o) = ZERO
    call rft( z2 )
    call rft_product( rft_y, z2 )
    call irft( z2 )

    call rft_product( rft_x, rft_v, z1, .true. )
    call irft( z1 )
    z1(n+1:o) = ZERO
    call rft( z1 )
    call rft_product( rft_x, z1 )
    call irft( z1 )
    
    w = z1(1:n) - z2(1:n)
    
    deallocate( rft_v, z1, z2, stat=info)
    if (info .ne. 0) then
       print *,'Error deallocating memory in gokh_sement'
       return
    end if      
 
    return
    
  end subroutine gokh_sement

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  ! If conjugated = .true. the first array is conjugated prior to the product
  subroutine rft_product( x, y, z_, conjugated_ )
    real(kind=r_kind), dimension(:),         intent(in)            :: x
    real(kind=r_kind), dimension(:), target, intent(inout)         :: y
    real(kind=r_kind), dimension(:), target, intent(out), optional :: z_
    logical,                                 intent(in),  optional :: conjugated_
   
    integer :: n, i
    real(kind=r_kind) :: x1, x2, y1, y2
    real(kind=r_kind), dimension(:), pointer :: z
    logical :: conjugated

    conjugated = .false.
    n = size(x,1)
    if( size(y,1).lt.n ) then
      print *,'error in rft_product: dimensions'
      return
    end if

    if( present(z_) ) then
      if( size(z_,1).lt.n ) then
        print *,'error in rft_product: dimensions'
        return
      end if
      z => z_
    else 
      z => y
    end if

    if( present(conjugated_) ) then
      if( conjugated_ ) conjugated = .true.
    end if

    if( .not.conjugated ) then
      z(1) = x(1) * y(1)
      do i=2,n-1,2
        x1     = x(i)
        x2     = x(i+1)
        y1     = y(i)
        y2     = y(i+1)
        z(i)   = x1*y1-x2*y2
        z(i+1) = x1*y2+x2*y1
      end do
      !We assume n is always even
      !if( mod(n,2).eq.0 ) z(n) = x(n) * y(n)
      z(n) = x(n) * y(n)
    else 
      z(1) = x(1) * y(1)
      do i=2,n-1,2
        x1     = x(i)
        x2     = -x(i+1)
        y1     = y(i)
        y2     = y(i+1)
        z(i)   = x1*y1-x2*y2
        z(i+1) = x1*y2+x2*y1
      end do
      !We assume n is always even
      !if( mod(n,2).eq.0 ) z(n) = x(n) * y(n)
      z(n) = x(n) * y(n)
    end if

  end subroutine rft_product

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! This routine applies the formula to a vector v
  ! to compute the product inv(T)*v, where T is Toeplitz
  ! Reentry routine for using in threads contexts.
  ! Initialization
  subroutine gokh_sement_init( x, rft_x, rft_y, alpha, y_ ) 
    !
    !     x   is the column of inv(T)
    !     y_  is the row of inv(T)
    !     rft_x  fft of the first term of the formula
    !     rft_y  fft of the second term of the formula
    !     alpha  =1/x(1)
    !
    real(kind=r_kind), dimension(:), intent(in)            :: x
    real(kind=r_kind), dimension(:), intent(in),  optional :: y_
    real(kind=r_kind), dimension(:), intent(out)           :: rft_x, rft_y
    real(kind=r_kind),               intent(out)           :: alpha

    real(kind=r_kind) :: eps

    logical :: non_symmetric
    
    integer :: n, o, info, i

    n = size(x,1)
    non_symmetric = present( y_ )
    o = size(rft_x,1)
    if( o.lt.n ) then
      print *,'gokh_sement_init: Dimensions error'
      return
    end if
    
    ! Testing the precision
    eps = epsilon(eps)
    if( eps*10000.gt.abs(x(1)) ) then
      print *,'gokh_sement_init warning: possible precision error (',x(1),')'
    end if

    if( non_symmetric ) then
      if( size(y_,1).ne.n ) then
        print *,'gokh_sement_init: Dimensions error'
        return
      end if
      print *,'gokh_sement_init: Non symmetric version still not implemented'
      return
    end if 
   
    alpha = ONE / ( x(1) * o * o )

    ! rft_x is the diagonal of the lower triangular factor of first member of the GS formula
    rft_x = (/x,(ZERO,i=1,o-n)/)
    call rft( rft_x )

    ! rft_y is the diagonal of the lower triangular factor of second member of the GS formula
    rft_y = (/ZERO,(x(n:2:-1)),(ZERO,i=1,o-n)/)
    call rft( rft_y )
    
  end subroutine gokh_sement_init

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  ! This routine applies the formula to a vector v
  ! to compute the product inv(T)*v, where T is Toeplitz
  ! Reentry routine for using in threads contexts.
  ! Application (only for non symmetric case)
  subroutine gokh_sement_apply( rft_x, rft_y, v, alpha, w_ ) 
    !
    !     rft_x  fft of the first term of the formula
    !     rft_y  fft of the second term of the formula
    !     v      is the array to be multiplied
    !     w_     is the array solution so v won't be written
    !     alpha  =1/x(1)
    !
    real(kind=r_kind), dimension(:), target, intent(inout)         :: v
    real(kind=r_kind), dimension(:), target, intent(out), optional :: w_
    real(kind=r_kind), dimension(:),         intent(in)            :: rft_x, rft_y
    real(kind=r_kind),                       intent(in)            :: alpha

    real(kind=r_kind), dimension(:), allocatable          :: rft_v, z1, z2
    real(kind=r_kind), dimension(:), pointer              :: w

    integer :: n, o, info, i

    n = size(v,1)
    o = size(rft_x,1)
    
    if( o.lt.n ) then
      print *,'gokh_sement_apply: Dimensions error'
      return
    end if

    if( present(w_) ) then
      w => w_
    else
      w => v
    end if
 
    allocate( rft_v(o), z1(o), z2(o), stat=info)
    if (info .ne. 0) then
       print *,'Error allocating memory in gokh_sement_apply'
       return
    end if      
 
    rft_v = alpha * (/v,(ZERO,i=1,o-n)/) 
    call rft( rft_v )
    call rft_product( rft_y, rft_v, z2, .true. )
    call irft( z2 )
    z2(n+1:o) = ZERO
    call rft( z2 )
    call rft_product( rft_y, z2 )
    call irft( z2 )

    call rft_product( rft_x, rft_v, z1, .true. )
    call irft( z1 )
    z1(n+1:o) = ZERO
    call rft( z1 )
    call rft_product( rft_x, z1 )
    call irft( z1 )
    
    w = z1(1:n) - z2(1:n)
    
    deallocate( rft_v, z1, z2, stat=info)
    if (info .ne. 0) then
       print *,'Error deallocating memory in gokh_sement_apply'
       return
    end if      
 
    return
    
  end subroutine gokh_sement_apply

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
end module gokh_sement_lib
