module rft_mkl
  use precision   ! exports kind value r_kind
  use mkl_dfti    ! mkl dft
  
  implicit none
  
  public :: rft, irft, rft_finalize
  private

  ! List data types
  type ll_rft_node
     integer :: n
     type(DFTI_DESCRIPTOR), pointer ::  Desc_Handle 
     type(ll_rft_node), pointer :: next => null()
  end type ll_rft_node

  type ll_rft_list
     type(ll_rft_node) :: head
  end type ll_rft_list

  type(ll_rft_list) :: ll_init_sizes


contains
  
  subroutine rft( w ) 
    
    !     rft performs a Real Fast Fourier Tranformation to a vector w. 
    !
    !     This routine uses the mlk.
    !     The Transformation is NOT normalized. A call to rft followed
    !     by a call to irft multiplies vector w by n
    !
    !     The first call to this routine initializes some vectors and workspace
    !     used in the subsequent calls with the same size. 
    !     A call with a different size produces a new initialization
    !
    !     This module must be compiled with openmp flag if will be use with
    !     conclurrent threads
    !
    real(kind=r_kind), dimension(:), intent(inout) :: w

    integer :: n, ierr
    
    n = size(w,1)
    
    !$OMP CRITICAL    
    if (.not. ll_rft_exists_n(ll_init_sizes, n)) then
      ierr = init_rft(n)
      if (ierr .ne. 0) stop 'Error initializing rft()'
    end if
    !$OMP END CRITICAL    
    
    call apply_rft(w,.true.)

    return
    
  end subroutine rft
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  

  subroutine irft( w ) 
    
    !     irft performs a Inverse Real Fast Fourier Tranformation to a vector w. 
    !
    !     This routine uses the mlk.
    !     The Transformation is NOT normalized. A call to rft followed
    !     by a call to irft multiplies vector w by n
    !
    !     The first call to this routine initializes some vectors and workspace
    !     used in the subsequent calls with the same size. 
    !     A call with a different size produces a new initialization
    !
    !     This module must be compiled with openmp flag if will be use with
    !     conclurrent threads
    !
    real(kind=r_kind), dimension(:), intent(inout) :: w

    integer :: n, ierr
    
    n = size(w,1)
    
    !$OMP CRITICAL    
    if (.not. ll_rft_exists_n(ll_init_sizes, n)) then
      ierr = init_rft(n)
      if (ierr .ne. 0) stop 'Error initializing rft()'
    end if
    !$OMP END CRITICAL    
    
    call apply_rft(w,.false.)

    return
    
  end subroutine irft
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  

  integer function init_rft(n) 
    integer, intent(in) :: n

    init_rft = ll_rft_add_n_rft(ll_init_sizes, n)

  end function init_rft

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  

  subroutine apply_rft(w,for_or_back) 
    real(kind=r_kind), dimension(:), intent(inout) :: w
    logical :: for_or_back

    integer :: n, info

    type(ll_rft_node), pointer :: p_node
   
    n = size(w,1)
    p_node => ll_rft_get_n_node(ll_init_sizes, n)
    if (.not. associated(p_node)) then
       print *, 'Error: trying to work with a non-initialized problem size'
       return
    end if

    if( for_or_back ) then 
      info = DftiComputeForward( p_node%Desc_Handle, w )
    else 
      info = DftiComputeBackward( p_node%Desc_Handle, w )
    end if
    call DftierrorManager( info )

  end subroutine apply_rft

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  

  integer function ll_rft_add_n_rft(ll, n)
    type(ll_rft_list), intent(inout) :: ll
    integer, intent(in) :: n

    type(ll_rft_node), pointer :: node

    integer :: info

    if (ll_rft_exists_n(ll, n)) then
       ll_rft_add_n_rft = 1
       return
    end if

    allocate(node, stat=info)
    if (info .ne. 0) then
       print *,'Error allocating memory in ll_rft_add_n_rft'
       ll_rft_add_n_rft = -1
    end if

    node%n = n

    info = DftiCreateDescriptor( node%Desc_Handle, DFTI_DOUBLE, DFTI_REAL, 1, node%n )
    call DftierrorManager( info )
    info = DftiSetValue( node%Desc_Handle, DFTI_PACKED_FORMAT, DFTI_PACK_FORMAT )
    call DftierrorManager( info )
    info = DftiCommitDescriptor( node%Desc_Handle )
    call DftierrorManager( info )

    node%next => ll%head%next
    ll%head%next => node

    ll_rft_add_n_rft = 0

  end function ll_rft_add_n_rft

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  logical function ll_rft_exists_n(ll, n)
    type(ll_rft_list), intent(in), target :: ll
    integer, intent(in) :: n

    type(ll_rft_node), pointer :: p_node
    
    ll_rft_exists_n = .false.

    p_node => ll%head
    do while (associated(p_node%next))
       p_node => p_node%next
       if (p_node%n .eq. n) then
          ll_rft_exists_n = .true.
          exit
       end if
    end do

  end function ll_rft_exists_n

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  function ll_rft_get_n_node(ll,n)
    type(ll_rft_node), pointer :: ll_rft_get_n_node
    type(ll_rft_list), intent(in), target :: ll
    integer, intent(in) :: n

    ll_rft_get_n_node => ll%head
    do while (associated(ll_rft_get_n_node%next))
       ll_rft_get_n_node => ll_rft_get_n_node%next
       if ( ll_rft_get_n_node%n .eq. n) then
          return
       end if
    end do

    nullify(ll_rft_get_n_node)

  end function ll_rft_get_n_node

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  

  subroutine DftierrorManager( Status )
    integer, intent(in) :: Status
    integer :: error
    if( DftiErrorClass( Status, error ) .eq. .true. ) then
      print *, DftiErrorMessage( Status )
      stop
    end if
  end subroutine DftierrorManager

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!  

  integer function rft_finalize()

    rft_finalize =  ll_rft_finalize(ll_init_sizes)

  end function rft_finalize

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  
  integer function ll_rft_finalize(ll)
    type(ll_rft_list), intent(inout), target :: ll
    integer :: i, n, ret
    
    n = ll_rft_length(ll)
    
    do i=1, n
       ret = ll_rft_remove_node(ll,1)
       if (ret .ne. 0) then
          ll_rft_finalize = -1
          return
       end if
    end do
    
    ll_rft_finalize = 0
    
  end function ll_rft_finalize
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function ll_rft_length(ll)
    type(ll_rft_list), intent(in), target :: ll
    type(ll_rft_node), pointer :: p_node
    integer :: n

    p_node => ll%head
    n=0
    do while (associated(p_node%next))
       p_node => p_node%next
       n = n+1
    end do

    ll_rft_length = n

  end function ll_rft_length

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  integer function ll_rft_remove_node(ll, i)
    type(ll_rft_list), intent(inout), target :: ll
    integer, intent(in) :: i

    type(ll_rft_node), pointer :: p_node, p_ant
    integer :: j, info
    
    p_ant  => ll%head
    p_node => ll%head%next
    do j=1,(i-1)
       p_ant => p_node
       p_node => p_node%next
    end do

    p_ant%next => p_node%next

    deallocate(p_node, stat=info)
    if (info .ne. 0) then
       print *,'Error deallocating memory in ll_rft_remove_node'
       ll_rft_remove_node = -2
    end if

    ll_rft_remove_node = 0

  end function ll_rft_remove_node
  
end module rft_mkl
