module toeplitz_cauchy
  use precision
  use dst_lib
  use diagCauchy
  use cauchy
  use cauchy_inertia
  use conv_lib
  implicit none
  
  public :: stoepsv, dtsyldl, dtsysv, dtsysv_neg, dtsysv_time
  real( kind = r_kind ) :: tiempo_dtsysv = ZERO
  private
  
contains
  
  subroutine stoepsv( T, B, X_, ref_ )
    !
    !   stoepsv solves a symmetric toeplitz system using the Cauchy method
    !
    real(kind=r_kind), dimension(:), intent(in) :: T 
    real(kind=r_kind), dimension(:), intent(inout), target :: B 
    real(kind=r_kind), dimension(:), intent(out), target, optional :: X_ 
    integer, intent(in), optional :: ref_ 

    real(kind=r_kind), dimension(:), pointer :: X
    real(kind=r_kind), parameter :: PI = FOUR * atan(ONE)
    real(kind=r_kind), dimension(size(T,1)) :: lambda, diag
    real(kind=r_kind), dimension(size(T,1)/2+mod(size(T,1),2)) :: g11, g12
    real(kind=r_kind), dimension(size(T,1)/2)                  :: g21, g22
    real(kind=r_kind), dimension(:,:), allocatable :: L1, L2
    real(kind=r_kind), dimension(size(T,1)) :: perm, x_aux, x1
    integer :: n, i, j, dim1, dim2, r
    integer :: ref = 0
    integer :: status

    n = size( T, 1 )
    if( present( ref_ ) ) ref = ref_

    ! Quick return if possible
    if (n .eq. 0) return  

    if( present(X_) ) then 
      X => X_
    else
      X => B
    end if

    ! Test input parameters
    if (n .lt. 4) then
       print *,'stoepsv error: dimension must be greater than 4.' 
       return
    end if
    
    if ( size(B,1) ) then
       print *,'stoepsv error: Dimension error.' 
       return
    end if
    
    if ( present(X_) ) then
      if( size(X_,1).ne.n ) then
        print *,'stoepsv error: Dimension error.' 
        return
      end if
    end if
    
    ! Initializing variables
    dim2 = n/2
    dim1 = dim2 + mod(n,2)
    allocate( L1(dim1,dim1), L2(dim2,dim2), Stat=status )
    if( status.ne.0 ) then 
      print *,'Error in tasks allocation( status = ',status,')'
      stop
    end if

    !Updating B for the Cauchy-like system
    perm = B
    call dst(perm)
    x_aux = (/perm(1:n:2),perm(2:n:2)/)/(TWO*(n+1))

    call cauchy_toep_generators(t,lambda(1:dim1),lambda(dim1+1:n),diag(1:dim1),diag(dim1+1:n), g11, g12, g21, g22 )

    !$OMP PARALLEL

    !$OMP SECTIONS 
    !$OMP SECTION

    ! LDL' decomposition of C1
    call cauchy_ldlt( lambda(1:dim1), diag(1:dim1), g11, g12, L1 ) 

    call dtrsv('l', 'n', 'u', dim1, L1, dim1, x_aux(1:dim1), 1)

    ! x_aux <-- inv(D)*x_aux
    do i=1,dim1
       x_aux(i) = x_aux(i) / L1(i,i)
    end do

    ! x_aux <-- L'\x_aux
    call dtrsv('l', 't', 'u', dim1, L1, dim1, x_aux(1:dim1), 1)

    !$OMP SECTION

    ! LDL' decomposition of C2
    call cauchy_ldlt( lambda(dim1+1:n), diag(dim1+1:n), g21, g22, L2 ) 

    !print *,'L2 = ',L2
    call dtrsv('l', 'n', 'u', dim2, L2, dim2, x_aux(dim1+1:n), 1)

    ! x_aux <-- inv(D)*x_aux
    j = 1
    do i=dim1+1,n
       x_aux(i) = x_aux(i) / L2(j,j)
       j = j + 1
    end do

    ! x_aux <-- L'\x_aux
    call dtrsv('l', 't', 'u', dim2, L2, dim2, x_aux(dim1+1:n), 1)

    !$OMP END SECTIONS

    !$OMP END PARALLEL

    !print *,'x_aux = ',x_aux
    x(1:n:2) = x_aux(1:dim1)
    x(2:n:2) = x_aux(dim1+1:n)
    call dst(x)

    do r = 1, ref
      x1 = x_aux
      !!!!!!!!!!!!!!!!!!
      !!! REFINEMENT !!!
      !!!!!!!!!!!!!!!!!!
      call conv( T, x, x_aux )
      !print *,'T*x = ',x_aux
      x_aux = B - x_aux
      !print *,'b-T*x = ',x_aux

      perm = x_aux
      call dst(perm)
      x_aux = (/perm(1:n:2),perm(2:n:2)/)/(TWO*(n+1))
      !print *,'dst(P(b-T*x)) = ',x_aux

      !!$OMP PARALLEL
      !!$OMP SECTIONS 
      !!$OMP SECTION

      call dtrsv('l', 'n', 'u', dim1, L1, dim1, x_aux(1:dim1), 1)

      ! x_aux <-- inv(D)*x_aux
      do i=1,dim1
         x_aux(i) = x_aux(i) / L1(i,i)
      end do

      ! x_aux <-- L'\x_aux
      call dtrsv('l', 't', 'u', dim1, L1, dim1, x_aux(1:dim1), 1)

      !!$OMP SECTION

      call dtrsv('l', 'n', 'u', dim2, L2, dim2, x_aux(dim1+1:n), 1)

      ! x_aux <-- inv(D)*x_aux
      j = 1
      do i=dim1+1,n
         x_aux(i) = x_aux(i) / L2(j,j)
         j = j + 1
      end do

      ! x_aux <-- L'\x_aux
      call dtrsv('l', 't', 'u', dim2, L2, dim2, x_aux(dim1+1:n), 1)

      !!$OMP END SECTIONS
      !!$OMP END PARALLEL


      !print *,'x1 = ',x1
      !x(1:n:2) =   x1(1:dim1) +   x_aux(1:dim1)
      !x(2:n:2) = x1(dim1+1:n) + x_aux(dim1+1:n)
      x_aux =   x1 + x_aux
      x(1:n:2) = x_aux(1:dim1)
      x(2:n:2) = x_aux(dim1+1:n)
      call dst(x)
    end do

    deallocate(L1,L2)

  end subroutine stoepsv

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine cauchy_toep_generators( t, lambda1, lambda2, diag1, diag2, g11, g12, g21, g22 )
    !
    !   cauchy_toep_generators obtains the rank displacement arguments of the two Cauchy-like 
    !   linear systems yielding from a symmetric Toeplitz matrix
    !
 
    !
    !   t: first column (row) of the toeplitz matrix
    !
    !   lambda1: diagonal of the displacement matrix of equation 1
    !
    !   lambda2: diagonal of the displacement matrix of equation 2
    !
    !   diag1: diagonal of the Cauchy-like matrix of equation 1
    !
    !   diag2: diagonal of the Cauchy-like matrix of equation 2
    !
    !   g11: First column of the generator of the displacement equation 1
    !
    !   g12: Second column of the generator of the displacement equation 1
    !
    !   g21: First column of the generator of the displacement equation 2
    !
    !   g22: Second column of the generator of the displacement equation 2
    !
    
    real(kind=r_kind), dimension(:), intent(in)  :: t
    real(kind=r_kind), dimension(:), intent(out) :: lambda1, diag1
    real(kind=r_kind), dimension(:), intent(out) :: g11, g12
    real(kind=r_kind), dimension(:), intent(out) :: lambda2, diag2
    real(kind=r_kind), dimension(:), intent(out) :: g21, g22

    integer :: n, n1, n2, i
    real(kind=r_kind), dimension(size(t,1)) :: perm, g1, g2, lambda, diag
    real(kind=r_kind) :: sqrtnp1, sqrtnp2
    real(kind=r_kind), parameter :: PI = FOUR * atan(ONE)
   
    n = size(t)
    n2 = n/2
    n1 = n2 + mod(n,2)

    sqrtnp1 = sqrt(TWO)
    sqrtnp2 = TWO * sqrtnp1
       
    !!$OMP PARALLEL private (i, perm)
    !!$OMP SECTIONS

    !!$OMP SECTION
    ! First column of generator G of T
    g1(1) = ZERO
    g1(2:n-1) = sqrtnp1 * T(3:n)
    g1(n) = ZERO
    call dst(g1)
    perm = g1
    g1 = (/perm(1:n:2),perm(2:n:2)/)
    g11 = g1(1:n1)
    g21 = g1(n1+1:n)

    !!$OMP SECTION
    ! Second column of generator G of C directly
    g2(1:n/2) = (/ (sqrtnp2*sin( real(i)*PI / real(n+1)),i=1,n/2) /)
    g2(n:(n+1)/2+1:-1) = g2( 1:n/2 )
    if (mod(n,2) .eq. 1) g2( n/2+1 ) = sqrtnp2
    perm = g2
    g2 = (/perm(1:n:2),perm(2:n:2)/)
    g12 = g2(1:n1)
    g22 = g2(n1+1:n)
    
    !!$OMP SECTION
    ! Computing vector lambda
    lambda(1:n/2) = (/ ( TWO * cos( real(i)*PI / real(n+1) ), i=1,n/2 ) /)
    lambda(n:(n+1)/2+1:-1) = -lambda( 1:n/2 )
    lambda = TWO * ( n + 1 ) * lambda
    if (mod(n,2) .eq. 1) lambda(n/2+1) = ZERO
    perm = lambda
    lambda = (/perm(1:n:2),perm(2:n:2)/)
    lambda1 = lambda(1:n1)
    lambda2 = lambda(n1+1:n)

    !!$OMP SECTION
    ! Computing diagonal entries of C a priori
    call diagC( t, diag )
    perm = diag
    diag = (/perm(1:n:2),perm(2:n:2)/)
    diag1 = diag(1:n1)
    diag2 = diag(n1+1:n)

  end subroutine cauchy_toep_generators

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine dtsyldl(T, L)
    ! ..Input/Output Arguments..
    real(kind=r_kind), intent(in) :: T(:)          ! Toeplitz column
    real(kind=r_kind), intent(out) :: L(:,:)       ! L(1:dim1,1:m) is the upper of the two triangular matrices of the decomp.
                                                   ! L(dim1+1:n,1:m) is the lower of the two triangular matrices of the decomp.
                                                   ! m = max(dim1,dim2)

    integer :: n, m, dim1, dim2
    real(kind=r_kind), dimension(size(T,1))                    :: lambda, diag
    real(kind=r_kind), dimension(size(T,1)/2+mod(size(T,1),2)) :: g11, g12
    real(kind=r_kind), dimension(size(T,1)/2)                  :: g21, g22

    n = size(T,1)
    m = size(L,2)
    if( size(L,1).lt.n .or. (m.lt.n/2+mod(n,2)) ) then 
      print *,'dtsyldl error'
      stop
    end if

    dim2 = n/2
    dim1 = dim2 + mod(n,2)

    call cauchy_toep_generators( T,lambda(1:dim1),lambda(dim1+1:n), &
                                   diag(1:dim1),diag(dim1+1:n),     & 
                                   g11, g12, g21, g22 )
    call cauchy_ldlt( lambda(1:dim1), diag(1:dim1), g11, g12, L(1:dim1,: ) )
    call cauchy_ldlt( lambda(dim1+1:n), diag(dim1+1:n), g21, g22, L(dim1+1:n,: ) )

  end subroutine dtsyldl

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! This routine solves a Toeplitz linear system by using 
  ! an already computed L factor. 
  ! It can be used refinement steps but it must passed 
  ! the Toeplitz column
  subroutine dtsysv( L, X, B, T_, ref_ )

    ! ..Input/Output Arguments..
    real(kind=r_kind), intent(in) :: L(:,:), B(:)
    real(kind=r_kind), intent(in), optional :: T_(:)
    real(kind=r_kind), intent(out) :: X(:)
    integer, intent(in), optional :: ref_ 

    real(kind=r_kind), dimension(size(X,1)) :: perm, x_aux, x1
    integer :: n, i, j, dim1, dim2, r
    integer :: ref = 0
    integer :: status

    real( kind=r_kind ) :: t1, t2, tn

    !print *,'dtsysv: '
    n = size( X, 1 )
    if( present( ref_ ) ) ref = ref_

    if( present(ref_).and.(.not.present(T_)) ) then
      print *,'Error in dtsysv arguments'
      stop
    end if

    ! Quick return if possible
    if (n .eq. 0) return  

    ! Test input parameters
    if (n .lt. 4) then
       print *,'dtsysv error: dimension must be greater than 4.' 
       return
    end if
    
    if ( (size(B,1).ne.n) .or. (size(L,1).ne.n) ) then
       print *,'dtsysv error: Dimension error.' 
       return
    end if
    
    ! Initializing variables
    dim2 = n/2
    dim1 = dim2 + mod(n,2)

    !Updating B for the Cauchy-like system
    perm = B
    call dst(perm)
    x_aux = (/perm(1:n:2),perm(2:n:2)/)/(TWO*(n+1))

    !$OMP PARALLEL

    !$OMP SECTIONS 
    !$OMP SECTION

    !print *,'dim1: ', dim1
    !print *,'L: ', L(1:dim1,:)
    !print *,'x_aux: ', x_aux(1:dim1)
    !print *,'n: ', n
!call ctimer( t1, tn, tn )
!print *,'t1'
    call dtrsv('l', 'n', 'u', dim1, L, n, x_aux, 1)
    !print *,'n: ', n
!call ctimer( t2, tn, tn )
!print *,'t2'
tiempo_dtsysv = tiempo_dtsysv + (t2-t1)

    ! x_aux <-- inv(D)*x_aux
    do i=1,dim1
       x_aux(i) = x_aux(i) / L(i,i)
    end do

    ! x_aux <-- L'\x_aux
    call dtrsv('l', 't', 'u', dim1, L, n, x_aux, 1)

    !$OMP SECTION

    !print *,'L2 = ',L2
    call dtrsv('l', 'n', 'u', dim2, L(dim1+1,1), n, x_aux(dim1+1), 1)

    ! x_aux <-- inv(D)*x_aux
    j = 1
    do i=dim1+1,n
       x_aux(i) = x_aux(i) / L(i,j)
       j = j + 1
    end do

    ! x_aux <-- L'\x_aux
    call dtrsv('l', 't', 'u', dim2, L(dim1+1,1), n, x_aux(dim1+1), 1)

    !$OMP END SECTIONS

    !$OMP END PARALLEL

    !print *,'x_aux = ',x_aux
    x(1:n:2) = x_aux(1:dim1)
    x(2:n:2) = x_aux(dim1+1:n)
    call dst(x)

    do r = 1, ref
      x1 = x_aux
      !!!!!!!!!!!!!!!!!!
      !!! REFINEMENT !!!
      !!!!!!!!!!!!!!!!!!
      call conv( T_, x, x_aux )
      !print *,'T*x = ',x_aux
      x_aux = B - x_aux
      !print *,'b-T*x = ',x_aux

      perm = x_aux
      call dst(perm)
      x_aux = (/perm(1:n:2),perm(2:n:2)/)/(TWO*(n+1))
      !print *,'dst(P(b-T*x)) = ',x_aux

      !!$OMP PARALLEL
      !!$OMP SECTIONS 
      !!$OMP SECTION

      call dtrsv('l', 'n', 'u', dim1, L(1:dim1,:), n, x_aux(1:dim1), 1)

      ! x_aux <-- inv(D)*x_aux
      do i=1,dim1
         x_aux(i) = x_aux(i) / L(i,i)
      end do

      ! x_aux <-- L'\x_aux
      call dtrsv('l', 't', 'u', dim1, L(1:dim1,:), n, x_aux(1:dim1), 1)

      !!$OMP SECTION

      call dtrsv('l', 'n', 'u', dim2, L(dim1+1:n,:), n, x_aux(dim1+1:n), 1)

      ! x_aux <-- inv(D)*x_aux
      j = 1
      do i=dim1+1,n
         x_aux(i) = x_aux(i) / L(i,j)
         j = j + 1
      end do

      ! x_aux <-- L'\x_aux
      call dtrsv('l', 't', 'u', dim2, L(dim1+1:n,:), n, x_aux(dim1+1:n), 1)

      !!$OMP END SECTIONS
      !!$OMP END PARALLEL


      !print *,'x1 = ',x1
      !x(1:n:2) =   x1(1:dim1) +   x_aux(1:dim1)
      !x(2:n:2) = x1(dim1+1:n) + x_aux(dim1+1:n)
      x_aux =   x1 + x_aux
      x(1:n:2) = x_aux(1:dim1)
      x(2:n:2) = x_aux(dim1+1:n)
      call dst(x)
    end do

  end subroutine dtsysv

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  ! Outputs the inertia of a symmetric Toeplitz matrix
  integer function dtsysv_neg( T )
    real(kind=r_kind), intent(in) :: T(:)          ! Toeplitz column

    integer :: n, dim1, dim2, neg
    real(kind=r_kind), dimension(size(T,1)) :: lambda, diag
    real(kind=r_kind), dimension(size(T,1)/2+mod(size(T,1),2)) :: g11, g12
    real(kind=r_kind), dimension(size(T,1)/2)                  :: g21, g22

    n = size(T,1)

    dim2 = n/2
    dim1 = dim2 + mod(n,2)

    call cauchy_toep_generators( T,lambda(1:dim1),lambda(dim1+1:n), &
                                   diag(1:dim1),diag(dim1+1:n),     & 
                                   g11, g12, g21, g22 )
    neg = cauchy_neg( lambda(1:dim1), diag(1:dim1), g11, g12 )
    neg = neg + cauchy_neg( lambda(dim1+1:n), diag(dim1+1:n), g21, g22 )

    dtsysv_neg = neg

  end function dtsysv_neg

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  function dtsysv_time
    real(kind=r_kind) :: dtsysv_time

    dtsysv_time=tiempo_dtsysv
  end function dtsysv_time

end module toeplitz_cauchy

