/***************************************************************************
 *   Copyright (C) 2008 by Leandro Graciá Gil                              *
 *   leagragi@fiv.upv.es                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 *  \file     cpu_eigen.c
 *  \brief    Symmetric Toeplitz matrix eigenvalue extractor (sequential CPU version)
 *  \author   Leandro Graciá Gil, leagragi@inf.upv.es
 *  \date     18/11/08
*/

// Standard library includes
#include <math.h>
#include <time.h>
#include <stdio.h>

// Main project header
#include "toeplitz.h"

// Command line parser info
#include "cmdline_cpu.h"

// Command line info structure
static struct gengetopt_args_info_cpu args_info;

/**
 * Parse command line commands into \a args_info structure (uses gengetopt).
 *
 * \param argc Number of params in command line.
 * \param argv Params of command line.
 * \param args_info Structure to be filled with parsed input params.
*/
void parse_cmdline(int argc, char *argv[], struct gengetopt_args_info_cpu *args_info) {

	if(cmdline_parser(argc, argv, args_info) != 0) {
		fprintf(stderr, "Run %s --help ('-h') to see the list of options.\n", argv[0]);
		exit(1);
	}
	
	if(args_info->help_given) {
		cmdline_parser_print_help();
		exit(0);
	}

	if(args_info->version_given) {
		cmdline_parser_print_version();
		exit(0);
	}
}

/**
 * CPU version of the Shift-and-invert 2-way Lanczos algorithm.
 *
 * \note Extracted eigenvalues are not sorted.
 *
 * \param n Size of \a t vector.
 * \param t Symmetric Toeplitz matrix data in vector form.
 * \param ia Start of the eigenvalue extraction interval.
 * \param ib End of the eigenvalue extraction interval.
 * \param sigma Shift value used for eigenvalue extraction.
 * \param nEig Number of required eigenvalues in the interval.
 * \param max_k Maximum size of Krylov subspace.
 * \param tolerance Tolerance threshold used to check eigenvalue convergence.
 * \param l_out Vector of size \a nEig containing extracted eigenvalues.
 * \param v_out Matrix of size \a n \c \a nEig (column order)containing extracted eigenvectors.
 * \return Number of eigenvalues extracted.
*/
int si2w_cpu(unsigned int n, const T *t, T ia, T ib, T sigma, unsigned int nEig, unsigned int max_k, T tolerance, T *l_out, T *v_out) {

	// Check no eigenvalue request
	if(nEig == 0) return 0;

	// Allocate memory for Lanczos vectors and tridiagonal matrix values
	T *a = (T *) malloc(max_k * sizeof(T));
	T *b = (T *) malloc(max_k * sizeof(T));
	T *g = (T *) malloc(max_k * sizeof(T));
	T *d = (T *) malloc(max_k * sizeof(T));
	T *p = (T *) malloc(n * max_k * sizeof(T));
	T *q = (T *) malloc(n * max_k * sizeof(T));
	T *pk = p, *qk = q;

	// Initialize symmetric and skew-symmetric vectors
	int i, j, k;
	memset(p, 0, n * sizeof(T));
	memset(q, 0, n * sizeof(T));
	p[0] = p[n-1] = (T)1.0;
	q[0] = (T)1.0; q[n-1] = (T)-1.0;

	// Normalize vectors
	T normP = _nrm2(n, p, 1);
	T normQ = _nrm2(n, q, 1);
	_scal(n, (T)1.0 / normP, p, 1);
	_scal(n, (T)1.0 / normQ, q, 1);

	// Calculate algorithm parameters
	int eigen_left = nEig;
	unsigned int extracted_s = 0, extracted_a = 0;
	unsigned int converged_s = 0, converged_a = 0;

	T ls[nEig], la[nEig];
	T *rs = (T *) malloc(nEig * n * sizeof(T));
	T *ra = (T *) malloc(nEig * n * sizeof(T));

	// Allocate required vectors
	T *w = (T *) malloc(n * sizeof(T));
	T *v = (T *) malloc(n * sizeof(T));
	T *vs = (T *) malloc(n * sizeof(T));
	T *va = (T *) malloc(n * sizeof(T));
	T *aux = (T *) malloc(n * sizeof(T));
	T *aux_k = (T *) malloc(max_k * sizeof(T));

	// Main loop
	for(k=1; k<=max_k; ++k) {

		// w = pk + qk
		for(i=0; i<n; ++i) w[i] = pk[i] + qk[i];

		// Solve (A - s * I) * v = w
		levinson(n, t[0] - sigma, &t[1], w, v);

		// Calculate vs and va vectors
		for(i=0; i<n; ++i) {
			vs[i] = (T)0.5 * (v[i] + v[n-i-1]);
			va[i] = (T)0.5 * (v[i] - v[n-i-1]);
		}

		// Dot product: ak = vs * pk, gk = va * qk
		a[k-1] = _dot(n, vs, 1, pk, 1);
		g[k-1] = _dot(n, va, 1, qk, 1);

		if(k < max_k) {
			// vs -= ak * pk, va -= gk * qk
			_axpy(n, -a[k-1], pk, 1, vs, 1);
			_axpy(n, -g[k-1], qk, 1, va, 1);

			// vs -= b(k-1) * p(k-1), va -= d(k-1) * q(k-1)
			if(k > 1) {
				_axpy(n, -b[k-2], pk - n, 1, vs, 1);
				_axpy(n, -d[k-2], qk - n, 1, va, 1);
			}

			// Reorthogonalization
			// vs -= p * (p' * vs)
			_gemv(CblasColMajor, CblasTrans, n, k, (T)1.0, p, n, vs, 1, 0.0, aux_k, 1);
			_gemv(CblasColMajor, CblasNoTrans, n, k, (T)-1.0, p, n, aux_k, 1, 1.0, vs, 1);

			// va -= q * (q' * va)
			_gemv(CblasColMajor, CblasTrans, n, k, (T)1.0, q, n, va, 1, 0.0, aux_k, 1);
			_gemv(CblasColMajor, CblasNoTrans, n, k, (T)-1.0, q, n, aux_k, 1, 1.0, va, 1);
		
			// bk = norm(vs), dk = norm(va)
			b[k-1] = _nrm2(n, vs, 1);
			d[k-1] = _nrm2(n, va, 1);

			// p(k+1) = vs / bk, q(k+1) = va / dk
			pk += n; qk += n;
			memset(pk, 0, n * sizeof(T));
			memset(qk, 0, n * sizeof(T));
			_axpy(n, (T)1.0 / b[k-1], vs, 1, pk, 1);
			_axpy(n, (T)1.0 / d[k-1], va, 1, qk, 1);
		}

		// Extract symmetric subspace eigenvalues
		eigen_left -= extract_eigenvalues(n, k, max_k, n, &converged_s, &extracted_s, eigen_left, ia, ib, t, p, a, b, tolerance, ls, rs);

		// Extract skew-symmetric subspace eigenvalues
		eigen_left -= extract_eigenvalues(n, k, max_k, n, &converged_a, &extracted_a, eigen_left, ia, ib, t, q, g, d, tolerance, la, ra);

		// Check convergence
		if(eigen_left <= 0) break;
	}

	// Merge eigenvalues and eigenvectors from both subspaces
	int l = 0;
	for(j=0; j<extracted_s; ++j) {
		l_out[l] = ls[j];
		memcpy(v_out + l * n, rs + j * n, n * sizeof(T));
		++l;
	}

	for(j=0; j<extracted_a; ++j) {
		l_out[l] = la[j];
		memcpy(v_out + l * n, ra + j * n, n * sizeof(T));
		++l;
	}

	// Free allocated memory resources
	free(w); free(v); free(va); free(vs);
	free(a); free(b); free(g); free(d);
	free(p); free(q); free(rs); free(ra);
	free(aux); free(aux_k);

	// Return number of eigenvalues extracted
	return l;
}

// Entry point
int main(int argc, char *argv[]) {

	// Parse command line input
	parse_cmdline(argc, argv, &args_info);

	// Check params: user-specified tolerance
	if(args_info.tolerance_arg <= 0.0f) {
		fprintf(stderr, "Error: invalid tolerance param value.\n");
		return 1;
	}

	// Check params: maximum number of eigenvalues per interval
	if(args_info.max_eigs_interval_arg <= 0) {
		fprintf(stderr, "Error: invalid value for max-eigs-interval param.\n");
		return 1;
	}

	// Check params: maximum subspace size
	if(args_info.max_k_arg <= 0) {
		fprintf(stderr, "Error: invalid value for max-k param.\n");
		return 1;
	}

	// Check params: maximum interval retries
	if(args_info.max_interval_retries_arg < 0) {
		fprintf(stderr, "Error: invalid value for max-interval-retries param.\n");
		return 1;
	}

	// Define algorithm parameters
	extract_global_info global;
	global.tolerance = args_info.tolerance_arg;
	global.max_eig_interval = args_info.max_eigs_interval_arg;
	global.max_k = args_info.max_k_arg;
	global.max_retry_count = args_info.max_interval_retries_arg;

	// Initialize random number generator
	if(args_info.random_seed_given) srand(args_info.random_seed_arg);
	else srand(time(NULL));
	unsigned int i;

	// Read input data
	if(args_info.file_given) {

		// Read from standard input
		if(!strcmp(args_info.file_arg, "-")) {
			if(!readInput((int *) &global.n, &global.t, stdin)) {
				fprintf(stderr, "Error reading data from standard input.\n");
				free(global.device);
				return 1;
			}
		
		// Read from a specified file
		} else {
			FILE *input = fopen(args_info.file_arg, "r");
			if(!input) {
				fprintf(stderr, "Error opening input file \"%s\"\n", args_info.file_arg);
				free(global.device);
				return 1;
			}

			if(!readInput((int *) &global.n, &global.t, input)) {
				fprintf(stderr, "Error reading input file \"%s\"\n", args_info.file_arg);
				free(global.device);
				fclose(input);
				return 1;
			}

			fclose(input);
		}
	}

	// Generate input data randomly
	else if(args_info.random_data_given) {

		// Check input data size
		if(args_info.random_data_arg <= 0) {
			fprintf(stderr, "Error: invalid data size.\n");
			free(global.device);
			return 1;
		}

		// Allocate memory
		global.n = args_info.random_data_arg;
		global.t = (T *) malloc(global.n * sizeof(T));

		// Fill input data randomly
		for(i=0; i<global.n; ++i) {
			global.t[i] = rand() / (T)RAND_MAX;
			global.t[i] *= (rand() & 1) ? (T)1 : (T)-1;
		}

		// Save randomly generated data
		if(args_info.save_random_data_given) {

			FILE *out = fopen(args_info.save_random_data_arg, "w");
			if(!out) {
				fprintf(stderr, "Error opening input file \"%s\"\n", args_info.save_random_data_arg);
				free(global.device);
				free(global.t);
				return 1;
			}

			fprintf(out, "%d\n", global.n);
			for(i=0; i<global.n; ++i) fprintf(out, "%.8f ", global.t[i]);
			fprintf(out, "\n");

			fclose(out);
		}
	}

	// Calculate Gershgorin disc interval
	T lbg, ubg;
	clock_t t1_intervals = clock();
	toeplitz_gershgorin(global.t, global.n, &lbg, &ubg);

	// Calculate eigenvalue intervals
	if(args_info.show_stats_flag) {
		printf("Calculating extraction intervals based on Gershgorin disc [%.6f, %.6f]: up to %d eigenvalues per interval.\n",
			lbg, ubg, global.max_eig_interval);
	}

	global.cpu = calc_eigen_intervals(global.t, global.n, lbg, ubg, global.max_eig_interval, &global.num_intervals);
	clock_t t2_intervals = clock();

	// Allocate memory for storing results
	T *out_l = (T *) malloc(global.n * sizeof(T));
	T *out_v = (T *) malloc(global.n * global.n * sizeof(T));
	T *out_l_int = out_l, *out_v_int = out_v;

	// Extract eigenvalues using shift and invert 2-way Lanczos
	clock_t t1_extraction = clock();
	unsigned int total_extracted = 0;
	unsigned int intervals_restarted = 0, restart_count = 0;
	unsigned int intervals_partial = 0, intervals_failed = 0;

	for(i=0; i<global.num_intervals; ++i) {
		interval_cpu_info *ic = &global.cpu[i];
		T sigma = (ic->lb + ic->ub) * (T)0.5;

		unsigned int extracted, retry = 0;
		do {
			extracted = si2w_cpu(global.n, global.t, ic->lb, ic->ub, sigma, ic->num_eigs, global.max_k, global.tolerance, out_l_int, out_v_int);
			if(extracted < ic->num_eigs && ++retry <= global.max_retry_count) {

				if(args_info.show_interval_status_flag) {
					printf("Retrying interval %d/%d: [%.6f, %.6f] with sigma %.6f, retry count = %d\n",
						i + 1, global.num_intervals, ic->lb, ic->ub, sigma, retry);
				}

				sigma = (rand() / (T)RAND_MAX) * (ic->ub - ic->lb) + ic->lb;

				if(retry == 1) ++intervals_restarted;
				++restart_count;
			}

			if(retry > global.max_retry_count) {
				if(args_info.show_interval_status_flag) {
					printf("Skipping interval %d/%d: [%.6f, %.6f] (max retry count reached, %d / %d eigenvalues extracted)\n",
						i + 1, global.num_intervals, ic->lb, ic->ub, ic->extracted_s + ic->extracted_a, ic->num_eigs);
				}

				if(extracted == 0) ++intervals_failed;
				else ++intervals_partial;

				break;
			}

		} while(extracted < ic->num_eigs);

		out_l_int += extracted;
		out_v_int += extracted * global.n;
		total_extracted += extracted;
	}

	clock_t t2_extraction = clock();

	// Sort eigenvalues and eigenvectors
	sort_eigenvalues(total_extracted, global.n, out_l, out_v);

	// Calculate time
	float time_intervals = (t2_intervals - t1_intervals) / (float)CLOCKS_PER_SEC;
	float time_extraction = (t2_extraction - t1_extraction) / (float)CLOCKS_PER_SEC;
	float time_total = time_intervals + time_extraction;

	// Show stats
	if(args_info.show_stats_flag) {
		printf("\n");

		printf("Extracted %d of %d (%.2f%%) eigenvalues (maximum subspace of size %d).\n",
			total_extracted, global.n, 100.0f * total_extracted / (float)global.n, global.max_k);
	
		printf("Total elapsed time: %.3fs. Interval precalculation: %.3fs (%.2f%%), eigenvalue extraction: %.3fs (%.2f%%).\n",
			time_total, time_intervals, 100.0f * time_intervals / time_total, time_extraction, 100.0f * time_extraction / time_total);

		printf("From a total of %d intervals, %d (%.2f%%) were restarted (%.2f retries/int), ",
			global.num_intervals, intervals_restarted, 100.0f * intervals_restarted / (float) global.num_intervals,
			restart_count / (float) max(1, intervals_restarted));

		printf("%d had partial extraction (%.2f%%) and %d failed to extract (%.2f%%).\n",
			intervals_partial, 100.0f * intervals_partial / (float) global.num_intervals, intervals_failed,
			100.0f * intervals_failed / (float) global.num_intervals);
	}
	
	// Show eigenvalues and eigenvectors
	if(args_info.show_eigenvalues_flag) print_vector("eigenvalues", out_l, total_extracted);
	if(args_info.show_eigenvectors_flag) print_matrix("eigenvectors", out_v, global.n, total_extracted, global.n);

	// Free memory
	free(global.t);
	free(global.cpu);
	free(out_l);
	free(out_v);

	return 0;
}

