/***************************************************************************
 *   Copyright (C) 2008 by Leandro Graciá Gil                              *
 *   leagragi@fiv.upv.es                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 *  \file     eigen.c
 *  \brief    Eigenvalue related functions: interval extraction, convergence checking and sorting.
 *  \author   Leandro Graciá Gil, leagragi@inf.upv.es
 *  \date     18/11/08
*/

// Main project header
#include "toeplitz.h"

// Auxiliar nested list structure used by fortran 90 code
typedef struct interval_node_struct {
	double lb, ub;
	integer eigs, prev_eigs, post_eigs;
	struct interval_node_struct *next;
} interval_node;

// External fortran 90 interval calculation function
void calc_intervals(interval_node *ilist, integer *n, double *t, double *lbg, double *ubg, integer *maxEig_interval);

// Auxiliar sorting structure
typedef struct sort_aux_struct_ {
	T el;
	unsigned int index;
} sort_aux_struct;

/**
 * Auxiliar quicksort comparison method for the \link ::sort_eigenvalues sort_eigenvalues \endlink fuction.
 * Check stdlib qsort documentation for details.
*/
static int sort_eigenvalues_aux(const void *p1, const void *p2) {
	T e1 = ((sort_aux_struct *)p1)->el;
	T e2 = ((sort_aux_struct *)p2)->el;

	if(e1 < e2) return -1;
	else if(e1 > e2) return 1;
	else return 0;
}

/**
 * Sort eigenvalues and eigenvectors using quicksort algorithm.
 *
 * \param extracted Number of extracted eigenvalues.
 * \param n Size of eigenvectors.
 * \param el Pointer to vector containing eigenvalues.
 * \param ev Pointer to matrix containing eigenvectors.
*/
void sort_eigenvalues(unsigned int extracted, unsigned int n, T *el, T *ev) {
	
	// Create an auxiliar sorting array
	unsigned int i;
	sort_aux_struct sort[extracted];
	for(i=0; i<extracted; ++i) {
		sort[i].el = el[i];
		sort[i].index = i;
	}

	// Sort auxiliar structure
	qsort(sort, extracted, sizeof(sort_aux_struct), sort_eigenvalues_aux);

	// Sort eigenvalues and eigenvectors
	T *ev_aux = (T *) malloc(extracted * n * sizeof(T));
	memcpy(ev_aux, ev, extracted * n * sizeof(T));

	for(i=0; i<extracted; ++i) {
		el[i] = sort[i].el;
		if(i != sort[i].index) memcpy(ev + i * n, ev_aux + sort[i].index * n, n * sizeof(T));
	}

	free(ev_aux);
}

/**
 * Calculates intervals containing approximately the same number of eigenvalues along a given interval.
 * Only interval-specific information is written in the array. All other data is initialized to 0.
 *
 * \param t Vector defining a symmetric Toeplitz matrix whose eigenvalue intervals will be calculated.
 * \param n Size of vector \a t.
 * \param lbg Lower bound of the initial interval to be subdivided (usually Gershgorin disc).
 * \param ubg Upper bound of the initial interval to be subdivided (usually Gershgorin disc).
 * \param max_eig_interval Maximum number of eigenvalues per interval.
 * \param num_intervals Number of extracted intervals.
 * \return Pointer to an array of size \a num_intervals containing interval info. \c NULL in case of error.
*/
interval_cpu_info *calc_eigen_intervals(T *t, unsigned int n, T lbg, T ubg, unsigned int max_eig_interval, unsigned int *num_intervals) {

	// Initialize interval count
	*num_intervals = 0;

	// Check params
	if(!t || !num_intervals || n == 0 || max_eig_interval == 0) return NULL;

	// Convert input vector to double precision, if required
	#ifdef SINGLE
	int i;
	double *t_double = (double *) malloc(n * sizeof(double));
	for(i=0; i<n; ++i) t_double[i] = t[i];
	#else
	double *t_double = t;
	#endif

	// Calculate eigenvalue extraction intervals
	integer n_ = n;
	interval_node interval_list;
	double lbg_ = lbg, ubg_ = ubg;
	calc_intervals(&interval_list, &n_, t_double, &lbg_, &ubg_, (integer *) &max_eig_interval);
	#ifdef SINGLE
	free(t_double);
	#endif

	// Count extracted intervals
	interval_node *i_node;
	for(i_node = interval_list.next; i_node; i_node = i_node->next, ++(*num_intervals));

	// Extract interval information
	interval_cpu_info *interval_cpu = (interval_cpu_info *) calloc(*num_intervals, sizeof(interval_cpu_info));
	interval_cpu_info *i_cpu = interval_cpu;

	for(i_node = interval_list.next; i_node; i_node = i_node->next, ++i_cpu) {

		// Store interval specific information
		i_cpu->lb = i_node->lb;
		i_cpu->ub = i_node->ub;
		i_cpu->num_eigs = i_node->eigs;
	}

	// Return interval information
	return interval_cpu;
}

/**
 * Check convergence and extract eigenvalues of a symmetric Toeplitz matrix given a tridiagonal shift-and-invert Lanczos approximation.
 *
 * \note sigma shift value is assumed to be the halfway between \a a and \a b.
 *
 * \param n Size of the Toeplitz matrix defined by \a t and number of rows of \a p.
 * \param k Current size of Krylov subspace and number of columns of \a p.
 * \param max_k Maximum allowed size of the Krylov subspace.
 * \param p_lda Leading dimension of Krylov subspace vectors stored in \a p.
 * \param converged Pointer to number of converged eigenvalues. Just initialize to \c 0 on start and let the function use it.
 * \param extracted Pointer to number of currently extracted eigenvalues.
 * \param required Number of eigenvalues that should be extracted.
 * \param a Start of the interval whose eigenvalues are being extracted.
 * \param b End of the interval whose eigenvalues are being extracted.
 * \param t Vector of size \a n defining the symmetric Toeplitz matrix.
 * \param p Matrix of size \a n x \a k containing Lanczos vectors.
 * \param alpha Vector of size \a k containing the main diagonal of the symmetric tridiagonal matrix.
 * \param beta Vector of size \a k-1 containing the second diagonal of the symmetric tridiagonal matrix.
 * \param tolerance Tolerance threshold used to check eigenvalue convergence.
 * \param l Output vector where up to \a required eigenvalues will be stored.
 * \param v Output matrix where up to \a required eigenvectors will be stored (column order).
 * \return Number of eigenvalues extracted during this call.
*/
unsigned int extract_eigenvalues(unsigned int n, unsigned int k, unsigned int max_k, unsigned int p_lda,
				 unsigned int *converged, unsigned int *extracted, unsigned int required,
				 T a, T b, const T *t, const T *p, const T *alpha, const T *beta, T tolerance,
				 T *l, T *v) {

	// Check params
	if(!converged || !extracted || !t || !p || !alpha || !beta || !l || !v) return 0;
	if(*extracted >= required) return 0;
	if(k > max_k) return 0;

	// Allocate temporary resources for eigenvalues and eigenvectors
	T *el = (T *) malloc(k * sizeof(T));
	T *ev = (T *) malloc(k * k * sizeof(T));
	T *work = (T *) malloc((2*k - 2) * sizeof(T));
	T *aux  = (T *) malloc(max(k, n) * sizeof(T));

	// Calculate eigenvalues and eigenvectors of the tridiagonal matrix
	{
		char jobz = 'V';
		integer n_ = k, ldz_ = k, info;

		memcpy(el, alpha, k * sizeof(T));
		memcpy(aux, beta, k * sizeof(T));

		_stev(&jobz, &n_, el, aux, ev, &ldz_, work, &info);
		if(info != 0) fprintf(stderr, "Error during tridiagonal eigenvalue extraction: %d elements did not converge\n", (int)info);
	}

	// Transform eigenvalues (shift-and-invert)
	int i, j;
	T sigma = (a + b) * (T)0.5;
	for(i=0; i<k; ++i) el[i] = sigma + (T)1.0 / el[i];

	// Calculate ritz vectors
	T *ritz = (T *) malloc(n * k * sizeof(T));
	_gemm(CblasColMajor, CblasNoTrans, CblasNoTrans, n, k, k, (T)1.0, p, p_lda, ev, k, (T)0.0, ritz, n);

	// Check eigenvalue convergence
	unsigned int extracted_now = 0;
	for(j=*converged; j<k; ++j) {
		if((j & 1) == 0) i = k - (j >> 1) - 1;
		else i = j >> 1;

		// Approximate residual (skip on last subspace iteration)
		if(k >= max_k || beta[k-1] * (ev + k * i)[k-1] <= tolerance) {

			// Calculate real residual if approximation seems to converge
			toeplitz_mult(n, t, ritz + i * n, aux);
			_axpy(n, -el[i], ritz + i * n, 1, aux, 1);
			T residual = _nrm2(n, aux, 1);

			// Process converged eigenvalue
			if(residual <= tolerance) {
				++(*converged);
				if(el[i] >= a && el[i] < b) {
					++extracted_now;
					l[*extracted] = el[i];
					memcpy(v + *extracted * n, ritz + i * n, n * sizeof(T));
					if(++(*extracted) >= required) break;
				}				
			}

		} else break;
	}
	
	// Release temporary resources
	free(el); free(ev); free(ritz);
	free(work); free(aux);

	// Return the number of eigenvalues extracted during this call
	return extracted_now;
}

