/***************************************************************************
 *   Copyright (C) 2008 by Leandro Graciá Gil                              *
 *   leagragi@fiv.upv.es                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 *  \file     gpu_eigen.cu
 *  \brief    Symmetric Toeplitz matrix eigenvalue extractor (multi-GPU version)
 *  \author   Leandro Graciá Gil, leagragi@inf.upv.es
 *  \date     18/11/08
*/

#include <stdio.h>
#include <assert.h>
#include <time.h>

// External C methods (required for CUDA linking)
extern "C" {

	// Main project header
	#include "toeplitz.h"
}

// Command line parser info
#include "cmdline_gpu.h"

// GPU kernel code
#include "toeplitz_kernel.cu"

// Command line info structure
static struct gengetopt_args_info_gpu args_info;

/**
 * Parse command line commands into \a args_info structure (uses gengetopt).
 *
 * \param argc Number of params in command line.
 * \param argv Params of command line.
 * \param args_info Structure to be filled with parsed input params.
*/
void parse_cmdline(int argc, char *argv[], struct gengetopt_args_info_gpu *args_info) {

	if(cmdline_parser(argc, argv, args_info) != 0) {
		fprintf(stderr, "Run %s --help ('-h') to see the list of options.\n", argv[0]);
		exit(1);
	}
	
	if(args_info->help_given) {
		cmdline_parser_print_help();
		exit(0);
	}

	if(args_info->version_given) {
		cmdline_parser_print_version();
		exit(0);
	}
}

/**
 * Assign a given interval to the current execution window.
 * Detaches assignment if an invalid interval number is given.
 *
 * \note GPU-side execution window should be updated once interval assignment finishes.
 *
 * \param global Global multi-GPU extraction structure describing eigenvalue extraction environment.
 * \param num_device Device number whose execution window should be used.
 * \param window_pos Execution window position to be assigned.
 * \param num_interval Index of the interval that should be assigned to execution window.
 * \param start_k Iteration number where execution is going to start.
*/
void assign_interval(extract_global_info *global, unsigned int num_device, unsigned int window_pos, unsigned int num_interval, unsigned int start_k) {

	// Check params
	if(!global) return;

	// Define a pointer to per-device info
	extract_device_info *info = &global->device[num_device];

	// Ignore invalid window positions
	if(window_pos >= info->window_size) return;

	// Detach interval if interval number is invalid
	if(num_interval >= global->num_intervals) {
		info->exec_window[window_pos].interval_index = INTERVAL_NONE;
		--info->window_used;
		--global->global_window_used;
		return;
	}

	// Declare pointers to interval to be assigned
	interval_cpu_info *cpu = &global->cpu[num_interval];
	interval_gpu_info *gpu = &global->gpu[num_interval];

	// Reset CPU extraction values
	cpu->extracted_s = 0; cpu->extracted_a = 0;
	cpu->converged_s = 0; cpu->converged_a = 0;
	cpu->eigen_left = cpu->num_eigs;

	// Configure GPU starting iteration
	gpu->start_k = start_k;
	gpu->processed = false;

	// Assign halfway point when initializing intervals and random values when relaunching
	if(info->exec_window[window_pos].interval_index != num_interval) gpu->sigma = (cpu->lb + cpu->ub) * (T)0.5;
	else gpu->sigma = (rand() / (T)RAND_MAX) * (cpu->ub - cpu->lb) + cpu->lb;

	// Assign CPU temporary memory
	cpu->p = info->cpu_p + window_pos * global->size_pq;
	cpu->q = info->cpu_q + window_pos * global->size_pq;
	cpu->m = info->cpu_m + window_pos * global->inc_k;

	cpu->alpha = info->cpu_alpha + window_pos * global->max_k;
	cpu->beta  = info->cpu_beta  + window_pos * global->max_k;
	cpu->gamma = info->cpu_gamma + window_pos * global->max_k;
	cpu->delta = info->cpu_delta + window_pos * global->max_k;
	
	// Assign GPU temporary memory
	gpu->p = info->gpu_p + window_pos * global->size_pq;
	gpu->q = info->gpu_q + window_pos * global->size_pq;
	gpu->m = info->gpu_m + window_pos * global->max_k;
	gpu->workspace = info->gpu_workspace + window_pos * global->size_workspace;

	// Increase execution window usage if there was no previous asignment
	if(info->exec_window[window_pos].interval_index == INTERVAL_NONE) {
		++info->window_used;
		++global->global_window_used;
	}

	// Asign execution window entry
	info->exec_window[window_pos].interval_index = num_interval;
}

/**
 * Releases all allocated memory for a multi-GPU eigenvalue extraction problem.
 *
 * \param global Global multi-GPU extraction structure describing eigenvalue extraction environment.
*/
void delete_extract_info(extract_global_info *global) {

	// Delete allocated information for each device
	for(unsigned int j=0; j<global->num_devices; ++j) {

		// Define a pointer to per-device info and select device
		extract_device_info *info = &global->device[j];
		cudaSetDevice(j);

		// Destroy stream objects
		cudaStreamDestroy(info->exec_stream);
		for(unsigned int i=0; i<info->window_size; ++i) cudaStreamDestroy(info->exec_window[i].copy_stream);

		// Free memory resources
		cudaFree(info->gpu_t);
		cudaFree(info->gpu_p);
		cudaFree(info->gpu_q);
		cudaFree(info->gpu_m);
		cudaFree(info->gpu_workspace);
		cudaFree(info->exec_window_gpu);

		cudaFreeHost(info->cpu_p);
		cudaFreeHost(info->cpu_q);
		cudaFreeHost(info->cpu_m);

		free(info->cpu_alpha);
		free(info->cpu_beta);
		free(info->cpu_gamma);
		free(info->cpu_delta);
		free(info->exec_window);
		free(info->device_prop);
	}

	for(unsigned int i=0; i<global->num_intervals; ++i) {
		free(global->cpu[i].ls);
		free(global->cpu[i].la);
		free(global->cpu[i].vs);
		free(global->cpu[i].va);
	}

	cudaFreeHost(global->gpu);
	free(global->cpu);
	free(global->device);
}

/**
 * Calculate required shared memory and maximum supported execution window size according to GPU limitations.
 * Window size written in the \a window_size attribute of the \a info parameter.
 *
 * \note Other \a info values such as \a size_pq, \a size_workspace and \a window_used are filled by this function.
 *
 * \param global Global multi-GPU extraction structure describing eigenvalue extraction environment.
 * \param max_window_size Maximum window size specified by the user. Set to \c 0 to use maximum possible size.
 * \return \c TRUE if enough global/shared memory to solve the problem. \c FALSE otherwise.
*/
int calc_memory_parameters(extract_global_info *global, unsigned int max_window_size) {

	// Check params
	if(!global) return FALSE;

	// Calculate some intermediate size values
	unsigned int n_align = align(global->n);
	
	global->size_pq = n_align * global->max_k;
	global->size_workspace = max(5 * n_align, 4 * n_align + global->max_k);
	global->size_toeplitz_data = (global->n - 1) * sizeof(T);
	global->size_smem = (6 * blockSize - 1) * sizeof(T);
	global->global_window_used = 0;

	int gmem_window_entry = (global->max_k * sizeof(tridiagonal_entry) + sizeof(interval_gpu_info) +
					 (2 * global->size_pq + global->size_workspace) * sizeof(T));

	// Check memory requirements for each device
	for(unsigned int i=0; i<global->num_devices; ++i) {

		// Define a pointer to per-device info and select device
		extract_device_info *info = &global->device[i];
		cudaSetDevice(i);

		// Allocate and get device properties
		info->device_prop = (cudaDeviceProp *) malloc(sizeof(cudaDeviceProp));
		if(cudaGetDeviceProperties(info->device_prop, i) != cudaSuccess) {
			fprintf(stderr, "Error obtaining device properties: %s\n", cudaGetErrorString(cudaGetLastError()));
			return 1;
		}

		// Check maximum number of registers per block
		if(blockSize > info->device_prop->maxThreadsPerBlock) {
			fprintf(stderr, "Fatal error: too many threads per block in device #%d. Try to reduce the blockSize macro.\n", i);
			return FALSE;
		}

		// Maximum window size is restricted by both global and shared memory limitations
		int max_window_size_gmem = (info->device_prop->totalGlobalMem - global->size_toeplitz_data) / gmem_window_entry;
		if(max_window_size_gmem <= 0) {
			fprintf(stderr, "Fatal error: not enough global memory in device #%d. Try to reduce maximum subspace size or number of eigenvalues per interval.\n", i);
			return FALSE;
		}

		// Note: 512 bytes of shared memory are left for extra shared variables in functions and argument passing
		int max_window_size_smem = info->device_prop->sharedMemPerBlock - global->size_smem - 512;
		if(max_window_size_smem <= 0) {
			fprintf(stderr, "Fatal error: not enough shared memory in device #%d. Try to reduce number of threads per block (blockSize macro).\n", i);
			return FALSE;
		}

		// Calculate execution window size
		info->window_size = min(max_window_size_gmem, max_window_size_smem);
		info->window_size = min(info->window_size, info->device_prop->maxGridSize[0]);
		info->window_size = min(info->window_size, global->num_intervals);
		info->window_used = 0;

		// Restrict size to user settings
		if(max_window_size > 0) info->window_size = min(info->window_size, max_window_size);
	}

	// Distribute uniformly execution window sizes
	int extra_slots = 0;
	int mean_slots = (global->num_intervals + global->num_devices - 1) / global->num_devices;
	int slots_left = global->num_intervals;

	for(unsigned int i=0; i<global->num_devices; ++i) {

		// Define a pointer to per-device info
		extract_device_info *info = &global->device[i];

		// Adjust window size
		extra_slots = max(0, mean_slots + extra_slots - (int)info->window_size);
		info->window_size = min(mean_slots + extra_slots, info->window_size);
		info->window_size = min(slots_left, info->window_size);
		slots_left -= info->window_size;

		// Calculate final memory requirements
		unsigned int gmem_size = info->window_size * gmem_window_entry + global->size_toeplitz_data;

		if(args_info.show_stats_flag) {
			printf("Device #%d: shared memory = %d/%d KB, global memory = %d/%d MB. Up to %d intervals evaluated in parallel.\n",
				i, global->size_smem / 1024, info->device_prop->sharedMemPerBlock / 1024, gmem_size / 1024 / 1024,
				info->device_prop->totalGlobalMem / 1024 / 1024, info->window_size);
		}
	}

	return TRUE;
}

/**
 * Allocate required memory and initialize execution window of each device.
 *
 * \param global Global multi-GPU extraction structure describing eigenvalue extraction environment.
*/
void prepare_devices(extract_global_info *global) {

	// Set interval asignment counter
	global->next_interval = 0;

	// Allocate page-locked memory for GPU-side interval data
	assert(cudaMallocHost((void **)&global->gpu, global->num_intervals * sizeof(interval_gpu_info)) != cudaErrorMemoryAllocation);

	// Allocate memory for extracted eigenvalues and eigenvectors
	for(unsigned int i=0; i<global->num_intervals; ++i) {
		global->cpu[i].ls = (T *) malloc(global->cpu[i].num_eigs * sizeof(T));
		global->cpu[i].la = (T *) malloc(global->cpu[i].num_eigs * sizeof(T));
		global->cpu[i].vs = (T *) malloc(global->cpu[i].num_eigs * global->n * sizeof(T));
		global->cpu[i].va = (T *) malloc(global->cpu[i].num_eigs * global->n * sizeof(T));
	}

	// For each CUDA device...
	for(unsigned int j=0; j<global->num_devices; ++j) {

		// Define a pointer to per-device info and select device
		extract_device_info *info = &global->device[j];
		cudaSetDevice(j);

		// Allocate required GPU memory for simultaneous execution
		assert(cudaMalloc((void **)&info->gpu_p, info->window_size * global->size_pq * sizeof(T)) != cudaErrorMemoryAllocation);
		assert(cudaMalloc((void **)&info->gpu_q, info->window_size * global->size_pq * sizeof(T)) != cudaErrorMemoryAllocation);
		assert(cudaMalloc((void **)&info->gpu_m, info->window_size * global->max_k * sizeof(tridiagonal_entry)) != cudaErrorMemoryAllocation);
		assert(cudaMalloc((void **)&info->gpu_workspace, info->window_size * global->size_workspace * sizeof(T)) != cudaErrorMemoryAllocation);
		assert(cudaMalloc((void **)&info->exec_window_gpu, info->window_size * sizeof(interval_gpu_info)) != cudaErrorMemoryAllocation);

		// Allocate required CPU memory and create asynchronous copy stream objects
		info->exec_window = (exec_window_entry *) malloc(info->window_size * sizeof(exec_window_entry));
		for(unsigned int i=0; i<info->window_size; ++i) {
			info->exec_window[i].interval_index = INTERVAL_NONE;
			cudaStreamCreate(&info->exec_window[i].copy_stream);
		}

		info->cpu_alpha = (T *) malloc(info->window_size * global->max_k * sizeof(T));
		info->cpu_beta  = (T *) malloc(info->window_size * global->max_k * sizeof(T));
		info->cpu_gamma = (T *) malloc(info->window_size * global->max_k * sizeof(T));
		info->cpu_delta = (T *) malloc(info->window_size * global->max_k * sizeof(T));

		// Buffers that exchange frequent information with GPU are page-locked for asynchronous transfers
		assert(cudaMallocHost((void **)&info->cpu_p, info->window_size * global->size_pq * sizeof(T)) != cudaErrorMemoryAllocation);
		assert(cudaMallocHost((void **)&info->cpu_q, info->window_size * global->size_pq * sizeof(T)) != cudaErrorMemoryAllocation);
		assert(cudaMallocHost((void **)&info->cpu_m, info->window_size * global->inc_k * sizeof(tridiagonal_entry)) != cudaErrorMemoryAllocation);

		// Initialize execution window
		for(unsigned int i=0; i<info->window_size; ++i) assign_interval(global, j, i, global->next_interval + i, 0);

		// Upload GPU-side execution window intervals to GPU memory
		cudaMemcpy(info->exec_window_gpu, global->gpu + global->next_interval, info->window_size * sizeof(interval_gpu_info), cudaMemcpyHostToDevice);
		global->next_interval += info->window_size;

		// Allocate and copy Toeplitz matrix data into GPU memory (main diagonal value is passed as an argument)
		assert(cudaMalloc((void **)&info->gpu_t, global->size_toeplitz_data) != cudaErrorMemoryAllocation);
		cudaMemcpy(info->gpu_t, &global->t[1], global->size_toeplitz_data, cudaMemcpyHostToDevice);

		// Create execution window stream object
		cudaStreamCreate(&info->exec_stream);
	}
}

// Entry point
int main(int argc, char *argv[]) {

	// Parse command line input
	parse_cmdline(argc, argv, &args_info);

	// Check presence of CUDA devices
	extract_global_info global;
	cudaGetDeviceCount((int *) &global.num_devices);

	if(global.num_devices == 0) {
		fprintf(stderr, "Fatal error: no CUDA capable devices found!!\n");
		return 1;
	}

	// Check params: user limitation of CUDA devices
	if(args_info.max_devices_given) {
		if(args_info.max_devices_arg <= 0) {
			fprintf(stderr, "Error: invalid value for max-devices param.\n");
			return 1;
		}

		global.num_devices = min(args_info.max_devices_arg, global.num_devices);
	}

	// Check params: user limitation of execution window size
	if(args_info.max_window_size_given && args_info.max_window_size_arg <= 0) {
		fprintf(stderr, "Error: invalid value for max-window-size param.\n");
		return 1;
	}

	// Check params: user-specified tolerance
	if(args_info.tolerance_arg <= 0.0f) {
		fprintf(stderr, "Error: invalid tolerance param value.\n");
		return 1;
	}

	// Check params: maximum number of eigenvalues per interval
	if(args_info.max_eigs_interval_arg <= 0) {
		fprintf(stderr, "Error: invalid value for max-eigs-interval param.\n");
		return 1;
	}

	// Check params: maximum subspace size
	if(args_info.max_k_arg <= 0) {
		fprintf(stderr, "Error: invalid value for max-k param.\n");
		return 1;
	}

	// Check params: subspace iteration size
	if(args_info.inc_k_arg <= 0) {
		fprintf(stderr, "Error: invalid value for inc-k param.\n");
		return 1;
	}

	// Check params: maximum interval retries
	if(args_info.max_interval_retries_arg < 0) {
		fprintf(stderr, "Error: invalid value for max-interval-retries param.\n");
		return 1;
	}

	// Define algorithm parameters
	global.tolerance = args_info.tolerance_arg;
	global.max_eig_interval = args_info.max_eigs_interval_arg;
	global.max_k = args_info.max_k_arg;
	global.inc_k = args_info.inc_k_arg;
	global.max_retry_count = args_info.max_interval_retries_arg;

	// Initialize per-device structure
	global.device = (extract_device_info *) calloc(global.num_devices, sizeof(extract_device_info));

	// Initialize random number generator
	if(args_info.random_seed_given) srand(args_info.random_seed_arg);
	else srand(time(NULL));

	// Read input data
	if(args_info.file_given) {

		// Read from standard input
		if(!strcmp(args_info.file_arg, "-")) {
			if(!readInput((int *) &global.n, &global.t, stdin)) {
				fprintf(stderr, "Error reading data from standard input.\n");
				free(global.device);
				return 1;
			}
		
		// Read from a specified file
		} else {
			FILE *input = fopen(args_info.file_arg, "r");
			if(!input) {
				fprintf(stderr, "Error opening input file \"%s\"\n", args_info.file_arg);
				free(global.device);
				return 1;
			}

			if(!readInput((int *) &global.n, &global.t, input)) {
				fprintf(stderr, "Error reading input file \"%s\"\n", args_info.file_arg);
				free(global.device);
				fclose(input);
				return 1;
			}

			fclose(input);
		}
	}

	// Generate input data randomly
	else if(args_info.random_data_given) {

		// Check input data size
		if(args_info.random_data_arg <= 0) {
			fprintf(stderr, "Error: invalid data size.\n");
			free(global.device);
			return 1;
		}

		// Allocate memory
		global.n = args_info.random_data_arg;
		global.t = (T *) malloc(global.n * sizeof(T));

		// Fill input data randomly
		for(unsigned int i=0; i<global.n; ++i) {
			global.t[i] = rand() / (T)RAND_MAX;
			global.t[i] *= (rand() & 1) ? (T)1 : (T)-1;
		}

		// Save randomly generated data
		if(args_info.save_random_data_given) {

			FILE *out = fopen(args_info.save_random_data_arg, "w");
			if(!out) {
				fprintf(stderr, "Error opening input file \"%s\"\n", args_info.save_random_data_arg);
				free(global.device);
				free(global.t);
				return 1;
			}

			fprintf(out, "%d\n", global.n);
			for(unsigned int i=0; i<global.n; ++i) fprintf(out, "%.8f ", global.t[i]);
			fprintf(out, "\n");

			fclose(out);
		}
	}

	// Calculate Gershgorin disc interval
	T lbg, ubg;
	clock_t t1_intervals = clock();
	toeplitz_gershgorin(global.t, global.n, &lbg, &ubg);

	// Calculate eigenvalue intervals
	if(args_info.show_stats_flag) {
		printf("Calculating extraction intervals based on Gershgorin disc [%.6f, %.6f]: up to %d eigenvalues per interval.\n",
			lbg, ubg, global.max_eig_interval);
	}

	global.cpu = calc_eigen_intervals(global.t, global.n, lbg, ubg, global.max_eig_interval, &global.num_intervals);
	clock_t t2_intervals = clock();

	// Calculate maximum execution window sizes according to global memory limitations
	unsigned int max_window_size = args_info.max_window_size_given ? args_info.max_window_size_arg : 0;
	if(!calc_memory_parameters(&global, max_window_size)) {
		for(unsigned int i=0; i<global.num_devices; ++i) free(global.device[i].device_prop);
		free(global.device);
		free(global.cpu);
		free(global.t);
		return 1;
	}

	// Allocate required memory and initialize execution windows of each device
	prepare_devices(&global);

	// Start time measurement
	unsigned int convergence_time_acc = 0;
	clock_t t1_extraction = clock();

	// Keep processing intervals until execution window is empty
	unsigned int k;
	for(k=0; global.global_window_used > 0; k += global.inc_k) {

		// Run all devices in parallel
		for(unsigned int dev=0; dev<global.num_devices; ++dev) {

			// Define a pointer to per-device info and select device
			extract_device_info *info = &global.device[dev];
			cudaSetDevice(dev);

			// Ensure that asynchronous transfer of execution window contents has finished
			cudaStreamSynchronize(info->exec_stream);

			// Run GPU 2-way Shift and Invert Lanczos in parallel, increasing subspaces up to info.max_k
			dim3 block(blockSize);
			dim3 grid(info->window_size);
			si2w_parallel <<< grid, block, global.size_smem >>> (global.n, global.t[0], info->gpu_t, k, global.inc_k, global.max_k, info->exec_window_gpu);
		}

		// Check convergence of results obtained from each device
		clock_t t1_convergence = clock();
		for(unsigned int dev=0; k > 0 && dev<global.num_devices; ++dev) {

			// Define a pointer to per-device info and select device
			extract_device_info *info = &global.device[dev];
			cudaSetDevice(dev);

			// For each increase of subspace done by the GPU...
			for(unsigned int j=0; j<global.inc_k; ++j) {

				// Calculate effective k
				unsigned int kj = k - global.inc_k + j;

				// Check convergence of each interval in execution window in parallel with GPU
				for(unsigned int i=0; i<info->window_size; ++i) {

					// Skip unused execution window slots
					int interval_index = info->exec_window[i].interval_index;
					if(interval_index == INTERVAL_NONE) continue;

					// Define information about current interval
					interval_gpu_info *ig = &global.gpu[interval_index];
					interval_cpu_info *ic = &global.cpu[interval_index];

					// Skip already processed intervals and new/just restarted intervals
					int k_real = kj - (int)ig->start_k;
					if(ig->processed || k_real < 0) continue;

					// Ensure that asynchronous transfers have finished
					cudaStreamSynchronize(info->exec_window[i].copy_stream);

					// Append tridiagonal entries to CPU interval data
					ic->alpha[k_real] = ic->m[j].alpha;
					ic->beta [k_real] = ic->m[j].beta;
					ic->gamma[k_real] = ic->m[j].gamma;
					ic->delta[k_real] = ic->m[j].delta;

					// Extract symmetric subspace eigenvalues from previous tridiagonal matrices
					unsigned int n_align = align(global.n);
					ic->eigen_left -= extract_eigenvalues(global.n, k_real + 1, global.max_k, n_align, &ic->converged_s,
								&ic->extracted_s, ic->eigen_left, ic->lb, ic->ub, global.t, ic->p, ic->alpha,
								ic->beta, global.tolerance, ic->ls, ic->vs);

					// Extract skew-symmetric subspace eigenvalues from previous tridiagonal matrices
					ic->eigen_left -= extract_eigenvalues(global.n, k_real + 1, global.max_k, n_align, &ic->converged_a,
								&ic->extracted_a, ic->eigen_left, ic->lb, ic->ub, global.t, ic->q, ic->gamma,
								ic->delta, global.tolerance, ic->la, ic->va);

					// Check convergence
					if(ic->eigen_left <= 0) {
	
						// Assign window slot to next free interval
						ig->processed = true;
						assign_interval(&global, dev, i, global.next_interval++, k + global.inc_k);
					}

					// Check if Krylov subspace has reached its maximum size
					else if(k_real + 1 >= global.max_k) {
	
						// Increase retry counter and check maximum value
						if(++ic->retry_count <= global.max_retry_count) {
							ig->sigma = (rand() / (T)RAND_MAX) * (ic->ub - ic->lb) + ic->lb;
							assign_interval(&global, dev, i, interval_index, k + global.inc_k);

							if(args_info.show_interval_status_flag) {
								printf("Retrying interval %d/%d in device #%d: [%.6f, %.6f] with sigma %.6f, retry count = %d\n",
									interval_index + 1, global.num_intervals, dev, ic->lb, ic->ub, ig->sigma, ic->retry_count);
							}

						} else {
							// Mark as processed and assign a new interval to the window slot
							assign_interval(&global, dev, i, global.next_interval++, k + global.inc_k);
							ig->processed = true;

							if(args_info.show_interval_status_flag) {
								printf("Skipping interval %d/%d in device #%d: [%.6f, %.6f] (max retry count reached, %d / %d eigenvalues extracted)\n",
									interval_index + 1, global.num_intervals, dev, ic->lb, ic->ub, ic->extracted_s + ic->extracted_a, ic->num_eigs);
							}
						}
					}
				}
			}
		}

		// Count time spent in checking eigenvalue convergence
		clock_t t2_convergence = clock();
		convergence_time_acc += t2_convergence - t1_convergence;

		// Check convergence of results obtained from each device
		for(unsigned int dev=0; dev<global.num_devices; ++dev) {

			// Define a pointer to per-device info and select device
			extract_device_info *info = &global.device[dev];
			cudaSetDevice(dev);

			// Synchronize with GPU
			cudaThreadSynchronize();

			// Check convergence
			if(info->window_used <= 0) continue;

			// Update interval data in GPU and copy current iteration results back to CPU
			for(unsigned i=0; i<info->window_size; ++i) {

				// Skip unused execution window slots
				int interval_index = info->exec_window[i].interval_index;
				if(interval_index == INTERVAL_NONE) continue;

				// Define information about current interval
				interval_gpu_info *ig = &global.gpu[interval_index];
				interval_cpu_info *ic = &global.cpu[interval_index];

				// Upload modified interval data to GPU memory
				int k_real = k - (int)ig->start_k;
				if(k_real < 0 || k_real + global.inc_k >= global.max_k) {
					cudaMemcpyAsync(info->exec_window_gpu + i, global.gpu + interval_index, sizeof(interval_gpu_info), cudaMemcpyHostToDevice, info->exec_stream);
				}

				// Skip already processed intervals and new/just restarted intervals
				if(ig->processed || k_real < 0) continue;

				// Get Lanczos vectors and tridiagonal entries asynchronously
				unsigned int n_align = align(global.n);
				unsigned int num_it = min(global.inc_k, global.max_k - k_real);
				cudaMemcpyAsync(ic->m, ig->m + k_real, num_it * sizeof(tridiagonal_entry), cudaMemcpyDeviceToHost, info->exec_window[i].copy_stream);
				cudaMemcpyAsync(ic->p + n_align * k_real, ig->p + n_align * k_real, num_it * n_align * sizeof(T), cudaMemcpyDeviceToHost, info->exec_window[i].copy_stream);
				cudaMemcpyAsync(ic->q + n_align * k_real, ig->q + n_align * k_real, num_it * n_align * sizeof(T), cudaMemcpyDeviceToHost, info->exec_window[i].copy_stream);
			}
		}
	}

	// Stop time measurement
	clock_t t2_extraction = clock();

	// Merge extracted eigenvalues in a vector and get interval restart statistics
	unsigned int total_extracted = 0;
	unsigned int intervals_restarted = 0, restart_count = 0;
	unsigned int intervals_partial = 0, intervals_failed = 0;

	for(unsigned int i=0; i<global.num_intervals; ++i) {
		unsigned int extracted = global.cpu[i].extracted_s + global.cpu[i].extracted_a;
		total_extracted += extracted;

		if(global.cpu[i].retry_count > 0) {
			restart_count += global.cpu[i].retry_count - 1;
			++intervals_restarted;

			if(global.cpu[i].retry_count > global.max_retry_count) {
				if(extracted == 0) ++intervals_failed;
				else ++intervals_partial;
			}
		}
	}

	T *el = (T *) malloc(total_extracted * sizeof(T));
	T *ev = (T *) malloc(total_extracted * global.n * sizeof(T));
	T *el_p = el, *ev_p = ev;

	for(unsigned int i=0; i<global.num_intervals; ++i) {
		interval_cpu_info *ic = &global.cpu[i];

		// Copy extracted eigenvalues
		memcpy(el_p, ic->ls, ic->extracted_s * sizeof(T));
		el_p += ic->extracted_s;

		memcpy(el_p, ic->la, ic->extracted_a * sizeof(T));
		el_p += ic->extracted_a;

		// Copy extracted eigenvectors
		memcpy(ev_p, ic->vs, ic->extracted_s * global.n * sizeof(T));
		ev_p += ic->extracted_s * global.n;

		memcpy(ev_p, ic->va, ic->extracted_a * global.n * sizeof(T));
		ev_p += ic->extracted_a * global.n;
	}

	// Release allocated memory
	delete_extract_info(&global);
	
	// Sort eigenvalues and eigenvectors
	sort_eigenvalues(total_extracted, global.n, el, ev);

	// Calculate time
	float time_intervals = (t2_intervals - t1_intervals) / (float)CLOCKS_PER_SEC;
	float time_extraction = (t2_extraction - t1_extraction) / (float)CLOCKS_PER_SEC;
	float time_convergence = convergence_time_acc / (float)CLOCKS_PER_SEC;
	float time_total = time_intervals + time_extraction;

	// Show stats
	if(args_info.show_stats_flag) {
		printf("\n");

		printf("Extracted %d of %d (%.2f%%) eigenvalues (maximum subspace of size %d).\n",
			total_extracted, global.n, 100.0f * total_extracted / (float)global.n, global.max_k);
	
		printf("Total elapsed time: %.3fs. Interval precalculation: %.3fs (%.2f%%), eigenvalue extraction: %.3fs (%.2f%%), convergence checking: %.3fs.\n",
			time_total, time_intervals, 100.0f * time_intervals / time_total, time_extraction, 100.0f * time_extraction / time_total, time_convergence);

		printf("From a total of %d intervals, %d (%.2f%%) were restarted (%.2f retries/int), ",
			global.num_intervals, intervals_restarted, 100.0f * intervals_restarted / (float) global.num_intervals,
			restart_count / (float) max(1, intervals_restarted));

		printf("%d had partial extraction (%.2f%%) and %d failed to extract (%.2f%%).\n",
			intervals_partial, 100.0f * intervals_partial / (float) global.num_intervals, intervals_failed,
			100.0f * intervals_failed / (float) global.num_intervals);
	}

	// Show eigenvalues and eigenvectors
	if(args_info.show_eigenvalues_flag) print_vector("eigenvalues", el, total_extracted);
	if(args_info.show_eigenvectors_flag) print_matrix("eigenvectors", ev, global.n, total_extracted, global.n);

	// Free result data
	free(el);
	free(ev);

	return 0;
}

