/***************************************************************************
 *   Copyright (C) 2008 by Leandro Graciá Gil                              *
 *   leagragi@fiv.upv.es                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 *  \file     gpu_vector_kernel.cu
 *  \brief    Kernel functions for vector operations.
 *  \author   Leandro Graciá Gil, leagragi@inf.upv.es
 *  \date     18/11/08
*/

// Main project header
#include "toeplitz.h"

/**
 * Calculates the 2-norm (length) of a given vector. 
 * Should be faster than sqrt(dot(x, x)) since reads each component only once.
 *
 * \param n Size of \a x vector.
 * \param x Input vector.
 * \param swork Shared memory workspace. Requires \c blockSize floats.
 * \return Thread 0 of the block returns the dot product result.
*/
__device__ T norm(unsigned int n, const T *x, T *swork) {

	// Allocate temporary blocks in shared memory and calculate thread id
	unsigned int tid = threadIdx.x;
	T *br = swork; // Temporary result block (size: blockSize)
	T result = (T)0.0;

	#ifdef SINGLE_THREAD
	// Process data using a single thread and without access patterns
	if(tid == 0) {
		for(unsigned int i=0; i<n; ++i) result += x[i] * x[i];
		#ifdef SINGLE
		result = sqrtf(result);
		#else
		result = sqrt(result);
		#endif
	}
	__syncthreads();
	return result;
	#endif

	// Process vector in blocks
	for(unsigned int b=0; b<n; b+=blockSize) {
		unsigned int i = b + tid;

		// Calculate per-block sum of squares
		if(i < n) {
			T aux = x[i];
			br[tid] = aux * aux;
		} else br[tid] = (T)0.0;
		__syncthreads();

		// Sum temporary result elements
		// Speed-up hack: some reductions are done in parallel, while small sets are reduced sequentially
		int s = blockSize;
		while(s > 8) {
			s >>= 1;
			if(tid < s) br[tid] += br[tid + s];
			__syncthreads();
		}

		if(tid == 0) {	
			for(unsigned int j=0; j<s; ++j)	result += br[j];
		}
		__syncthreads();
	}

	if(tid == 0) {
		#ifdef SINGLE
		result = sqrtf(result);
		#else
		result = sqrt(result);
		#endif
	}
	__syncthreads();

	// Thread 0 returns the correct value
	return result;
}


/**
 * Calculates the dot product of \a x and \a y vectors.
 *
 * \param n Size of \a x and \a y vectors.
 * \param x First vector.
 * \param y Second vector.
 * \param swork Shared memory workspace. Requires \c blockSize floats.
 * \return Thread 0 of the block returns the dot product result.
*/
__device__ T dot(unsigned int n, const T *x, const T *y, T *swork) {

	// Allocate temporary blocks in shared memory and calculate thread id
	unsigned int tid = threadIdx.x;
	T *br = swork; // Temporary result block (size: blockSize)
	T result = (T)0.0;

	#ifdef SINGLE_THREAD
	// Process data using a single thread and without access patterns
	if(tid == 0) {
		for(unsigned int i=0; i<n; ++i) result += x[i] * y[i];
	}
	__syncthreads();
	return result;
	#endif

	// Process vector in blocks
	for(unsigned int b=0; b<n; b+=blockSize) {
		unsigned int i = b + tid;

		// Calculate per-block dot product
		if(i < n) br[tid] = x[i] * y[i];
		else br[tid] = (T)0.0;
		__syncthreads();

		// Sum temporary result elements
		// Speed-up hack: some reductions are done in parallel, while small sets are reduced sequentially
		int s = blockSize;
		while(s > 8) {
			s >>= 1;
			if(tid < s) br[tid] += br[tid + s];
			__syncthreads();
		}

		if(tid == 0) {	
			for(unsigned int j=0; j<s; ++j)	result += br[j];
		}
		__syncthreads();
	}

	// Thread 0 returns the correct value
	return result;
}

/**
 * Calculates the dot product of \a x and reverse(\a y) vectors.
 *
 * \param n Size of \a x and \a y vectors.
 * \param x First vector.
 * \param y Second vector.
 * \param swork Shared memory workspace. Requires \c 4 * \c blockSize - 1 floats.
 * \return Thread 0 of the block returns the dot product result.
*/
__device__ T dot_reverse_y(unsigned int n, const T *x, const T *y, T *swork) {
	
	// Allocate temporary blocks in shared memory
	T *bx = swork;                   // Forward vector block (size: 2 * blockSize - 1)
	T *by = bx + blockSize * 2 - 1;  // Backward vector block (size: blockSize)
	T *br = by + blockSize;          // Temporary result block (size: blockSize)

	// Calculate thread id and aligned vector size
	unsigned int tid = threadIdx.x;
	unsigned int n_align = align(n);
	unsigned int shift = n_align - n;
	T result = 0.0;
	
	#ifdef SHOW_PARAMS
	// Show function parameters (for debug purposes)
	if(tid == 0) {
		fprintf(stderr, "dot_reverse_y: n = %d, x = [", n);
		for(int i=0; i<n; ++i) fprintf(stderr, " %.6f", x[i]);
		fprintf(stderr, " ], y = [");
		for(int i=0; i<n; ++i) fprintf(stderr, " %.6f", y[i]);
		fprintf(stderr, " ]\n");
	}
	#endif

	#ifdef SINGLE_THREAD
	// Process data using a single thread and without access patterns
	if(tid == 0) {
		for(unsigned int i=0; i<n; ++i) result += x[i] * y[n-i-1];
	}
	__syncthreads();
	return result;
	#endif

	// Reset shifted contents of forward vector
	bx[tid] = (T)0.0;
	__syncthreads();

	// Process vectors in blocks
	for(unsigned int i=0; i<n; i+=blockSize) {
		
		// Read forward and reverse blocks in parallel into shared memory
		unsigned int xi = i + tid;
		unsigned int yi = n_align - i - blockSize + tid;

		if(xi < n) bx[shift + tid] = x[xi];
		if(yi < n) by[tid] = y[yi]; else by[tid] = (T)0.0;
		__syncthreads();
	
		// Calculate per-block dot product
		br[tid] = bx[tid] * by[blockSize - tid - 1];
		__syncthreads();

		// Shift contents of forward block
		if(tid < shift) bx[tid] = bx[tid + blockSize];

		// Sum temporary result elements
		// Speed-up hack: some reductions are done in parallel, while small sets are reduced sequentially
		int s = blockSize;
		while(s > 8) {
			s >>= 1;
			if(tid < s) br[tid] += br[tid + s];
			__syncthreads();
		}

		if(tid == 0) {	
			for(unsigned int j=0; j<s; ++j)	result += br[j];
		}
		__syncthreads();
	}

	// Thread 0 returns the correct value
	return result;
}

/**
 * Performs \a y = \a a *\a x + \a y operation.
 *
 * \param n Size of \a x and \a y vectors.
 * \param a Scalar factor.
 * \param x First vector.
 * \param y Second vector, modified with result.
*/
__device__ void axpy(unsigned int n, T a, const T *x, T *y) {
	
	// Blocksize should be aligned to half_warp
	unsigned int tid = threadIdx.x;

	#ifdef SINGLE_THREAD
	// Process data using a single thread and without access patterns
	if(tid == 0) {
		for(unsigned int i=0; i<n; ++i) y[i] += a * x[i];
	}
	__syncthreads();
	return;
	#endif

	// Process vectors in blocks
	for(unsigned int b=0; b<n; b+=blockSize) {
		
		// Calculate per-block axpy (all memory accesses should be aligned)
		unsigned int i = b + tid;
		if(i < n) y[i] += a * x[i];
	}
	__syncthreads();
}

/**
 * Performs \a y = \a a * reverse(\a x) + \a y operation.
 *
 * \param n Size of \a x and \a y vectors.
 * \param a Scalar factor.
 * \param x First vector.
 * \param y Second vector, modified with result. Cannot be \a x.
 * \param swork Shared memory workspace. Requires \c 5 * \c blockSize - 1 floats.
*/
__device__ void axpy_reverse_x(unsigned int n, T a, const T *x, T *y, T *swork) {

	// Allocate temporary blocks in shared memory
	T *bx = swork;                   // Backward vector block (size: blockSize)
	T *by = bx + blockSize;          // Forward vector block (size: 2 * blockSize - 1)
	T *br = by + 2 * blockSize - 1;  // Temporary result block (size: 2 * blockSize)

	// Calculate thread id and aligned vector size
	unsigned int tid = threadIdx.x;
	unsigned int n_align = align(n);
	unsigned int shift = n_align - n;
	unsigned int lastblock = blockSize - shift;

	#ifdef SHOW_PARAMS
	// Show function parameters (for debug purposes)
	if(tid == 0) {
		fprintf(stderr, "axpy_reverse_x: n = %d, a = %.6f, x = [", n, a);
		for(unsigned int i=0; i<n; ++i) fprintf(stderr, " %.6f", x[i]);
		fprintf(stderr, " ], y = [");
		for(unsigned int i=0; i<n; ++i) fprintf(stderr, " %.6f", y[i]);
		fprintf(stderr, " ]\n");
	}
	#endif

	#ifdef SINGLE_THREAD
	// Process data using a single thread and without access patterns
	if(tid == 0) {
		for(unsigned int i=0; i<n; ++i) y[i] += a * x[n-i-1];
	}
	__syncthreads();
	return;
	#endif

	// Process vectors in blocks
	for(unsigned int i=0; i<n; i+=blockSize) {
		
		// Read forward and reverse blocks in parallel into shared memory
		unsigned int xi = n_align - i - blockSize + tid;
		unsigned int yi = i + tid;

		if(xi < n) bx[tid] = x[xi];
		if(yi < n) by[shift + tid] = y[yi];
		__syncthreads();
	
		#ifdef DEBUG
		if(tid == 0) {
			fprintf(stderr, "by:");
			for(unsigned int i=0; i<blockSize*2-1; ++i) fprintf(stderr, " %.6f", by[i]);
			fprintf(stderr, "\n");

			fprintf(stderr, "bx:");
			for(unsigned int i=0; i<blockSize; ++i) fprintf(stderr, " %.6f", bx[i]);
			fprintf(stderr, "\n");
		}
		#endif
	
		// Calculate per-block axpy
		if(i > 0) {
			// Append to previous temporary results
			br[tid + lastblock] = a * bx[blockSize - tid - 1] + by[tid];

		} else {
			// Skip shift gap and put contents at block start
			if(tid >= shift) br[tid - shift] = a * bx[blockSize - tid - 1] + by[tid];
		}
		__syncthreads();

		#ifdef DEBUG
		if(tid == 0) {
			fprintf(stderr, "br:");
			for(unsigned int i=0; i<2*blockSize; ++i) fprintf(stderr, " %.6f", br[i]);
			fprintf(stderr, "\n\n");
		}
		#endif
	
		// Store temporary results
		if(i > 0) y[yi - blockSize] = br[tid];

		// Shift contents of forward block
		if(tid < shift) by[tid] = by[tid + blockSize];

		// Shift contents of result block
		if(i > 0 && tid < lastblock) br[tid] = br[tid + blockSize];
		__syncthreads();
	}

	// Store remaining contents of temporary result block
	unsigned int yi = n_align - blockSize + tid;
	if(yi < n) y[yi] = br[tid];
	__syncthreads();
}

/**
 * Performs \a x = \a b * (\a a * reverse(\a x) + \a x) operation.
 *
 * \param n Size of \a x vector.
 * \param a Scalar factor.
 * \param x Input vector.
 * \param y Output vector. Can be \a x.
 * \param b Result scaling value.
 * \param swork Shared memory workspace. Requires \c 6 * \c blockSize - 1 floats.
*/
__device__ void axpxb_reverse_x(unsigned int n, T a, const T *x, T *y, T b, T *swork) {

	// Allocate temporary blocks in shared memory
	T *bb = swork;                    // Backward vector block (size: blockSize)
	T *bf = bb + blockSize;           // Forward vector block (size: 2 * blockSize - 1)
	T *brf = bf + blockSize * 2 - 1;  // Temporary forward result block (size: blockSize * 2)
	T *brb = brf + blockSize * 2;     // Temporary backward result block (size: blockSize)

	// Calculate thread id and aligned vector size
	unsigned int tid = threadIdx.x;
	unsigned int n_align = align(n);
	unsigned int shift = n_align - n;
	unsigned int lastblock = blockSize - shift;

	#ifdef SHOW_PARAMS
	// Show function parameters (for debug purposes)
	if(tid == 0) {
		fprintf(stderr, "axpxb_reverse_x: n = %d, a = %.6f, x = [", n, a);
		for(unsigned int i=0; i<n; ++i) fprintf(stderr, " %.6f", x[i]);
		fprintf(stderr, " ]\n");
	}
	#endif

	#ifdef SINGLE_THREAD
	// Process data using a single thread and without access patterns
	if(tid == 0) {
		for(unsigned int i=0; i<n/2; ++i) {
			T aux = x[i];
			y[i] = b * (aux + a * x[n-i-1]);
			y[n-i-1] = b * (x[n-i-1] + a * aux);
		}

		if(n & 1) y[n/2] = b * (x[n/2] + a * x[n/2]);
	}
	__syncthreads();
	return;
	#endif

	// Process vectors in blocks
	unsigned int num_blocks = n_align / blockSize;
	unsigned int max_n = blockSize * (1 + (num_blocks >> 1));
	for(unsigned int i=0; i<max_n; i+=blockSize) {
		
		// Read forward and reverse blocks in parallel into shared memory
		unsigned int bi = n_align - i - blockSize + tid;
		unsigned int fi = i + tid;

		if(bi < n) bb[tid] = x[bi];
		if(fi < n) bf[shift + tid] = x[fi];
		__syncthreads();

		#ifdef DEBUG
		if(tid == 0) {
			fprintf(stderr, "bf:");
			for(unsigned int i=0; i<blockSize*2-1; ++i) fprintf(stderr, " %.6f", bf[i]);
			fprintf(stderr, "\n");

			fprintf(stderr, "bb:");
			for(unsigned int i=0; i<blockSize; ++i) fprintf(stderr, " %.6f", bb[i]);
			fprintf(stderr, "\n");
		}
		#endif
	
		// Calculate per-block axpxb
		if(i > 0) {
			// Append to previous temporary results
			brf[tid + lastblock] = b * (a * bb[blockSize - tid - 1] + bf[tid]);

			// Backward results doesn't need any appending
			brb[tid] = b * (bb[tid] + a * bf[blockSize - tid - 1]);

		} else {
			// Skip shift gap and put contents at block start
			if(tid >= shift) brf[tid - shift] = b * (a * bb[blockSize - tid - 1] + bf[tid]);

			// Store backward results at block start
			if(tid < lastblock) brb[tid] = b * (bb[tid] + a * bf[blockSize - tid - 1]);
		}
		__syncthreads();

		#ifdef DEBUG
		if(tid == 0) {
			fprintf(stderr, "brf:");
			for(unsigned int i=0; i<2*blockSize; ++i) fprintf(stderr, " %.6f", brf[i]);
			fprintf(stderr, "\nbrb:");
			for(unsigned int i=0; i<blockSize; ++i) fprintf(stderr, " %.6f", brb[i]);
			fprintf(stderr, "\n\n");
		}
		#endif

		// Store temporary results from forward result block
		if(i > 0) y[fi - blockSize] = brf[tid];

		// Skip writing last backward result in last iteration if number of blocks is even
		if(i + blockSize >= max_n && !(num_blocks & 1)) break;

		// Store temporary results from backward result block
		if(bi < n) y[bi] = brb[tid];
		
		// Avoid unnecessary shifting in last iteration
		if(i + blockSize >= max_n) break;

		// Shift contents of forward block
		if(tid < shift) bf[tid] = bf[tid + blockSize];

		// Shift contents of forward result block
		if(i > 0 && tid < lastblock) brf[tid] = brf[tid + blockSize];
		__syncthreads();
	}
	__syncthreads();
}

