/***************************************************************************
 *   Copyright (C) 2008 by Leandro Graciá Gil                              *
 *   leagragi@fiv.upv.es                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 *  \file     toeplitz.c
 *  \brief    Toeplitz matrix related functions.
 *  \author   Leandro Graciá Gil, leagragi@inf.upv.es
 *  \date     18/11/08
*/

// Standard library includes
#include <math.h>

// Main project header
#include "toeplitz.h"

/**
 * Calculate Gershgorin disc for a given Toeplitz symmetric matrix.
 * 
 * \param t Vector defining a symmetric Toeplitz matrix.
 * \param n Size of vector \a t.
 * \param low Lower bound of the Gershgorin disc interval.
 * \param up Upper bound of the Gershgorin disc interval.
*/
void toeplitz_gershgorin(const T *t, unsigned int n, T *low, T *up) {

	// Check params
	if(!t || !low || !up) return;

	// Calculate Gershgorin disc interval
	int i;
	T sum = 0.0;
	for(i=1; i<n; ++i) sum += fabs(t[i]);

	T rad = sum;
	for(i=1; i<(n+1)/2; ++i) {
		sum = sum - fabs(t[n-i]) + fabs(t[i]);
		if(sum > rad) rad = sum;
	}

	*low = t[0] - rad;
	*up  = t[0] + rad;
}

/**
 * Calculate y = toeplitz(t) * x matrix-vector product where t is in vector form.
 * 
 * \param n Size of \a t and \a x, \a y vectors.
 * \param t Data of the symmetric Toeplitz matrix in vector form.
 * \param x Data of the input vector.
 * \param y Data of the destination vector. Will be overwritten with the result.
*/
void toeplitz_mult(int n, const T *t, const T *x, T *y) {

	// Check params
	if(n <= 0) return;

	// Calculate contribution of main diagonal
	memset(y, 0, n * sizeof(T));
	_axpy(n, t[0], x, 1, y, 1);
	
	// Add contribution of each subdiagonal
	unsigned int i;
	for(i=1; i<n; ++i) {
		_axpy(n-i, t[i], x + i, 1, y, 1);
		_axpy(n-i, t[i], x, 1, y + i, 1);
	}
}

/**
 * Solve system toeplitz(r0) * x = b0 where r0 defines a symmetric positive-definite Toeplitz matrix.
 *
 * \param n Size of \a r0, \a b0 and \a x vectors.
 * \param td Main diagonal element of the symmetric Toeplitz matrix.
 * \param t0 Rest of the symmetric Toeplitz matrix data in vector form.
 * \param b0 Right-hand side vector data.
 * \param x Result vector.
 * \note If matrix is not positive-definite the algorithm can still converge.
*/
void levinson(unsigned int n, T td, const T *t0, const T *b0, T *x) {

	// b = b0 / td;
	T inv_td = (T)1.0 / td;
	T *b = (T *) malloc(n * sizeof(T));
	memcpy(b, b0, n * sizeof(T));
	_scal(n, inv_td, b, 1);

	// t = t0(2:n) / td;
	T *t = (T *) malloc((n-1) * sizeof(T));
	memcpy(t, t0, (n-1) * sizeof(T));
	_scal(n-1, inv_td, t, 1);

	T *y = (T *) malloc(n * sizeof(T));
	x[0] = b[0];
	y[0] = -t[0];

	T alpha = -t[0];
	T beta = (T)1.0;

	T *aux = (T *) malloc(n * sizeof(T));

	int k;
	for(k=1; k<n; ++k) {
		beta = ((T)1.0 - alpha * alpha) * beta;
		T mu = (b[k] - _dot(k, t, 1, x, -1)) / beta;
		_axpy(k, mu, y, -1, x, 1);
		x[k] = mu;

		if(k < n-1) {
			alpha = -(t[k] + _dot(k, t, 1, y, -1)) / beta;

			// y(1:k) += alpha * y(k:-1:1);
			memcpy(aux, y, k * sizeof(T));
			_axpy(k, alpha, aux, -1, y, 1);
			y[k] = alpha;
		}
	}

	free(b);
	free(t);
	free(y);
	free(aux);
}

