/***************************************************************************
 *   Copyright (C) 2008 by Leandro Graciá Gil                              *
 *   leagragi@fiv.upv.es                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 *  \file     toeplitz.h
 *  \brief    Main project header file.
 *  \author   Leandro Graciá Gil, leagragi@inf.upv.es
 *  \date     18/11/08
*/

#ifndef _TOEPLITZ_H_
#define _TOEPLITZ_H_

// Standard library includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// Blas and lapack includes
#include <f2c.h>
#include <clapack.h>
#include <gsl/gsl_cblas.h>

// CUDA includes
#include <cuda_runtime.h>

// Precision options
#define SINGLE            // Use simple precision floating point format instead of double

// CUDA kernel debug options
//#define SHOW_PARAMS     // Show parameters of critical functions
//#define SINGLE_THREAD   // Use only one tread and don't care about memory alignment (safest results)
//#define DEBUG           // Show intermediate calculations in some functions

// Definitions for CUDA kernels
// Note: blockSize should not be lower than 64 for correct coalescing
#define blockSize 256 //512
#define align(x) (((x) + 255) & ~255) //(((x) + 511) & ~511)

#define half_warp 16
#define half_warp_mask 0x0F

#define print_vector_gpu(name, v, n) { if(tid == 0) { \
					fprintf(stderr, "%s: [", name); \
					for(int _i=0; _i<n; ++_i) fprintf(stderr, " %.6f", v[_i]); \
					fprintf(stderr, " ]\n"); } }

// Definitions for blas and lapack functions
#ifdef SINGLE
	#define T float
	#define Tsf "%f"
	#define _nrm2 cblas_snrm2
	#define _scal cblas_sscal
	#define _axpy cblas_saxpy
	#define _gemv cblas_sgemv
	#define _gemm cblas_sgemm
	#define _spmv cblas_sspmv
	#define _dot  cblas_sdot
	#define _spsv sspsv_
	#define _stev sstev_
#else
	#define T double
	#define Tsf "%lf"
	#define _nrm2 cblas_dnrm2
	#define _scal cblas_dscal
	#define _axpy cblas_daxpy
	#define _gemv cblas_dgemv
	#define _gemm cblas_dgemm
	#define _spmv cblas_dspmv
	#define _dot  cblas_ddot
	#define _spsv dspsv_
	#define _stev dstev_
#endif

// Define boolean values, if not already done
#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

// Auxiliar definition values
#define INTERVAL_NONE -1

// 2-way tridiagonal matrix entry structure
#ifdef __GNUC__
typedef struct __attribute__ ((aligned(16 * sizeof(T)))) tridiagonal_entry_struct {
#else
typedef struct __align__(16 * sizeof(T)) tridiagonal_entry_struct {
#endif
	T alpha;
	T gamma;
	T beta;
	T delta;
} tridiagonal_entry;

// Interval CPU-side information
typedef struct interval_cpu_info_struct {

	// Interval specific information
	T lb, ub;                               ///< Lower and upper bounds of the interval.
	unsigned int num_eigs;                  ///< Number of eigenvalues in the interval.

	// Eigenvalue extraction information
	T *p, *q;                               ///< Symmetric and skew-symmetric Lanczos subspace bases.
	T *ls, *la, *vs, *va;                   ///< Eigenvalues and eigenvectors extracted from symmetric and skew-symmetric subspaces.
	T *alpha, *beta, *gamma, *delta;        ///< Tridiagonal entries from symmetric and skew-symmetric subspaces.	
	tridiagonal_entry *m;                   ///< Tridiagonal entries from symmetric and skew-symmetric subspaces. Used for asynchronous copies.

	int eigen_left;                         ///< Number of eigenvalues remaining to be extracted.
	unsigned int extracted_s, extracted_a;  ///< Number of eigenvalues already extracted from symmetric and skew-symmetric subspaces.
	unsigned int converged_s, converged_a;  ///< Number of already converged eigenvalues from symmetric and skew-symmetric subspaces.
	unsigned int retry_count;               ///< Number of times that the interval has been retried.

} interval_cpu_info;

// Interval GPU-side information (aligned to 16 * sizeof(float) size to ensure coalescing)
#ifdef __GNUC__
typedef struct __attribute__ ((aligned(16 * sizeof(float)))) interval_gpu_info_struct {
#else
typedef struct __align__(16 * sizeof(float)) interval_gpu_info_struct {
#endif
	unsigned int start_k;     ///< Iteration in which interval extraction was launched.
	unsigned char processed;  ///< Flag indicating if interval has been already processed.
	T sigma;                  ///< Shift value used in the interval.
	T *p, *q;                 ///< Matrices of size n_align x max_k. Contains base vectors for symmetric and skew-symmetric Krylov subspaces.
	T *workspace;             ///< Vector of size max(5 * align(\a n), 4 * align(\a n) + \a max_k).
	tridiagonal_entry *m;     ///< Struct vector of size max_k.
} interval_gpu_info;

// Execution window slot information
typedef struct exec_window_entry_struct {
	int interval_index;       ///< Associated interval index or \c INTERVAL_NONE.
	cudaStream_t copy_stream; ///< Asynchronous copy stream object.
} exec_window_entry;

// Per-device extraction information
typedef struct extract_device_info_struct {

	// Execution window information
	unsigned int window_size;            ///< Number of intervals executed simultaneously.
	unsigned int window_used;            ///< Number of active intervals in the execution window.
	exec_window_entry *exec_window;      ///< Indices of intervals in execution. Set to \c INTERVAL_NONE if no interval associated.
	interval_gpu_info *exec_window_gpu;  ///< Pointers to interval descriptors stored in global GPU memory.
	cudaStream_t exec_stream;            ///< Asynchronous copy stream object for execution window data.

	// Device specific information
	struct cudaDeviceProp *device_prop;  ///< CUDA device properties.

	// Auxiliar memory information
	T *gpu_t;                            ///< GPU array storing problem data (main diagonal not included, passed as an argument).
	T *gpu_p, *gpu_q;                    ///< GPU arrays for Krylov subspace matrices (for all execution window intervals).
	T *gpu_workspace;                    ///< GPU arrays for auxiliar workspace (for all execution window intervals).
	tridiagonal_entry *gpu_m;            ///< GPU arrays for storing tridiagonal results (for all execution window intervals).

	T *cpu_p, *cpu_q;                    ///< CPU arrays for Krylov subspace matrices (for all execution window intervals).
	T *cpu_alpha, *cpu_beta;             ///< CPU arrays for symmetric subspace tridiagonal results (for all execution window intervals).
	T *cpu_gamma, *cpu_delta;            ///< CPU arrays for skew-symmetric subspace tridiagonal results (for all execution window intervals).
	tridiagonal_entry *cpu_m;            ///< CPU arrays for skew-symmetric subspace tridiagonal results (for all execution window intervals).

} extract_device_info;

// Global multi-GPU extraction information
typedef struct extract_global_info_struct {

	// Global size parameters
	unsigned int num_intervals;          ///< Number of total intervals.
	unsigned int max_eig_interval;       ///< Maximum number of eigenvalues per interval.
	unsigned int max_k;                  ///< Maximum Krylov subspace size in all intervals.
	unsigned int max_retry_count;        ///< Maximum of retries of each interval.
	unsigned int inc_k;                  ///< Subspace size increment in each iteration.

	// Problem parameters
	unsigned int n;                      ///< Size of the problem.
	T *t;                                ///< Problem data values defining a symmetric Toeplitz matrix.
	T tolerance;                         ///< Required tolerance for eigenvalue extraction.

	// Global internal information
	interval_cpu_info *cpu;              ///< CPU-side interval information.
	interval_gpu_info *gpu;              ///< GPU-side interval information (stored in CPU memory).
	unsigned int next_interval;          ///< Next interval to be processed.
	unsigned int global_window_used;     ///< Number of active intervals in execution windows of all devices.

	// Per-device information
	unsigned int num_devices;            ///< Number of CUDA-enabled devices available.
	extract_device_info *device;         ///< Per-device eigenvalue extraction information.

	// Auxiliar calculated sizes
	unsigned int size_pq;                ///< Size, in T elements, of Krylov subspace matrices for each interval.
	unsigned int size_workspace;         ///< Size, in T elements, of each interval's workspace.
	unsigned int size_toeplitz_data;     ///< Size, in bytes, of problem data defining a symmetric Toeplitz matrix.
	unsigned int size_smem;              ///< Required preallocated shared memory by GPU kernels.

} extract_global_info;

// Function prototypes from utils.c
int  readInput(int *n, T **t, FILE *file);
void print_vector(const char *name, T *v, int n);
void print_matrix(const char *name, T *m, int cols, int rows, int lda);

// Function prototypes from toeplitz.c
void toeplitz_gershgorin(const T *t, unsigned int n, T *low, T *up);
void toeplitz_mult(int n, const T *t, const T *x, T *y);
void levinson(unsigned int n, T td, const T *t0, const T *b0, T *x);

// Function prototypes from eigen.c
void sort_eigenvalues(unsigned int extracted, unsigned int n, T *el, T *ev);
interval_cpu_info *calc_eigen_intervals(T *t, unsigned int n, T lbg, T ubg, unsigned int max_eig_interval, unsigned int *num_intervals);
unsigned int extract_eigenvalues(unsigned int n, unsigned int k, unsigned int max_k, unsigned int p_lda,
				 unsigned int *converged, unsigned int *extracted, unsigned int required,
				 T a, T b, const T *t, const T *p, const T *alpha, const T *beta, T tolerance,
				 T *l, T *v);
#endif

