/***************************************************************************
 *   Copyright (C) 2008 by Leandro Graciá Gil                              *
 *   leagragi@fiv.upv.es                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**
 *  \file     toeplitz_kernel.cu
 *  \brief    Kernel functions for Levinson algorithm and shift and invert 2-way Lanczos.
 *  \author   Leandro Graciá Gil, leagragi@inf.upv.es
 *  \date     18/11/08
*/

// Main project header
#include "toeplitz.h"

// Vector operations kernel
#include "gpu_vector_kernel.cu"

/**
 * Perform a coalesced raw copy from aligned addresses with data sizes multiple of 4.
 *
 * \param dest Source address. Assumed to be aligned to 16 * sizeof(float) if points to global memory.
 * \param src Destination address. Assumed to be aligned to 16 * sizeof(float) if points to global memory.
 * \param size Data size in bytes. Should be always a multiple of 4.
*/
__device__ void memcpy4(void *dest, void *src, unsigned int size) {

	unsigned int tid = threadIdx.x;
	float *destf = (float *)dest;
	float *srcf  = (float *)src;
	
	__syncthreads();

	size >>= 2;
	for(unsigned int b=0; b<size; b+=blockSize, srcf+=blockSize, destf+=blockSize) {
		unsigned int i = b + tid;
		if(i < size) destf[tid] = srcf[tid];
	}
	
	__syncthreads();
}

/**
 * Solve a symmetric Toeplitz system of the form toeplitz(t) * x = b
 *
 * \note This function requires \a blockSize >= \a half_warp.
 *
 * \param n Size of \a [td t], \a x and \a y vectors (size of Toeplitz matrix).
 * \param td Diagional value of the symmetric Toeplitz matrix.
 * \param t Vector defining the rest of symmetric Toeplitz matrix values.
 * \param b Right-hand side of the system. Will be normalized by \a td.
 * \param x Vector that will contain the solution of the system.
 * \param work Work vector of size 2 * align(\a n).
 * \param swork Shared memory workspace. Requires \c 6 * \c blockSize - 1 floats.
*/
__device__ void levinson(unsigned int n, T td, const T *t, T *b, T *x, T *work, T *swork) {

	// Calculate thread identifier and define block size
	unsigned int tid = threadIdx.x;
	unsigned int n_align = align(n);

	#ifdef SHOW_PARAMS
	// Show function parameters (for debug purposes)
	if(tid == 0) {
		fprintf(stderr, "levinson: n = %d, td = %.6f, t = [", n, td);
		for(int i=0; i<n-1; ++i) fprintf(stderr, " %.6f", t[i]);
		fprintf(stderr, " ], b = [");
		for(int i=0; i<n; ++i) fprintf(stderr, " %.6f", b[i]);
		fprintf(stderr, " ]\n");
	}
	#endif

	// Check parameters
	if(n == 0) return;

	// Assign workspace
	T *y  = work;
	T *tn = y + n_align;

	// Initialize auxiliar variables
	__shared__ T b_cache[half_warp], tn_cache[half_warp];
	__shared__ T alpha, mu;	
	T beta = (T)1.0;

	// Divide t elements by td in parallel
	T factor = (T)1.0 / td;
	for(unsigned int i=0; i<n-1; i+=blockSize) {
		unsigned int ti = i + tid;
		if(ti < n-1) {
			T aux = t[ti] * factor;
			if(ti < half_warp) tn_cache[ti] = aux;
			tn[ti] = aux;
		}

		__syncthreads();
	}

	// Divide b elements by td in parallel
	for(unsigned int i=0; i<n; i+=blockSize) {
		volatile unsigned int bi = i + tid;
		if(bi < n) {
			T aux = b[bi] * factor;
			if(bi < half_warp) b_cache[bi] = aux;
			b[bi] = aux;
		}

		__syncthreads();
	}

	// Initialize x, y elements
	if(tid == 0) {
		x[0] = b_cache[0];
		y[0] = -tn_cache[0];
		alpha = -tn_cache[0];
	}
	__syncthreads();

	// Main loop
	for(int k=1; k<n; ++k) {

		// Reload b & tn vector caches in parallel
		unsigned int k_half_warp = k & half_warp_mask;
		if(k_half_warp == 0) {
			if(tid < half_warp) {
				unsigned int i = k + tid;
				if(i < n) {
					b_cache[tid] = b[i];
					tn_cache[tid] = tn[i];
				}
			}
			__syncthreads();
		}

		// Scalar values are processed by thread 0
		T dot_result = dot_reverse_y(k, tn, x, swork);
		if(tid == 0) {
			beta = ((T)1.0 - alpha * alpha) * beta;
			mu = (b_cache[k_half_warp] - dot_result) / beta;
		}
		__syncthreads();

		axpy_reverse_x(k, mu, y, x, swork);
		if(tid == k_half_warp) x[k] = mu;
		__syncthreads();

		if(k < n-1) {
			dot_result = dot_reverse_y(k, tn, y, swork);
			if(tid == 0) alpha = -(tn_cache[k_half_warp] + dot_result) / beta;
			__syncthreads();

			axpxb_reverse_x(k, alpha, y, y, (T)1.0, swork);
			if(tid == k_half_warp) y[k] = alpha;
		}

		__syncthreads();
	}
}

/**
 * Reorthogonalize a vector to a Krylov subspace base.
 *
 * \note This function requires \a blockSize >= \a half_warp.
 *
 * \param n Size of \a v vector and number of rows of \a p matrix.
 * \param k_size Current size of Krylov subspace. Number of columns of \a p matrix.
 * \param p Pointer to a column-ordered matrix defining the base of the Krylov subspace.
 * \param v Pointer to the vector to be reorthogonalized.
 * \param h Work array of size \a k_size.
 * \param swork Shared memory workspace. Requires \c 2 * \c blockSize floats.
*/
__device__ void reorthogonalize(unsigned int n, unsigned int k_size, const T *p, T *v, T *h, T *swork) {

	// Calculate thread identifier and define block size
	unsigned int tid = threadIdx.x;
	unsigned int n_align = align(n);

	// Calculate h = p' * v
	const T *pk = p;
	for(unsigned int bk=0; bk<k_size; bk+=half_warp) {

		// Temporary result block for dot products
		__shared__ T dr[half_warp];

		// Calculate h in blocks aligned to optimal memory access
		for(unsigned int w=0; w<half_warp && bk+w<k_size; ++w, pk+=n_align) {

			// Calculate k-th element of h = p' * v 
			T aux = dot(n, pk, v, swork);
			if(tid == 0) dr[w] = aux;
			__syncthreads();
		}

		// Store h data block
		unsigned int bi = bk + tid;
		if(tid < half_warp && bi < k_size) h[bi] = dr[tid];
		__syncthreads();
	}

	// Calculate vs -= p * h
	for(unsigned int bn=0; bn<n; bn+=blockSize) {
		unsigned int j = bn + tid;
		
		// Allocate temporary blocks in shared memory
		T *bh = swork;          // Temporary block for aligned read of h contents (size: blockSize)
		T *ba = bh + blockSize; // Temporary result accumulation block (size: blockSize)

		// Clear accumulator block
		ba[tid] = (T)0.0;
		__syncthreads();

		// For each k-block of columns...	
		for(unsigned int bk=0; bk<k_size; bk+=blockSize) {
			unsigned int k = bk + tid;

			// Read h data into temporary block
			if(k < k_size) bh[tid] = h[k];
			__syncthreads();

			// Use k-th element to scale k-th column of p and add to accumulator
			for(unsigned int ci=0; ci<k_size; ++ci) {
				if(j < n) ba[tid] += bh[ci] * p[ci * n_align + j];
			}
		}
		__syncthreads();

		// Substract accumulator block results to v vector
		if(j < n) v[j] -= ba[tid];
	}

	__syncthreads();
}

/**
 * Shift and invert 2-way Lanczos routine for Toeplitz matrices.
 * Calculates symmetric and skew-symmetric tridiagonal values, but leaves final eigenvalue calculation to CPU side.
 *
 * \note This function requires \a blockSize >= \a half_warp.
 *
 * \param n      Size of the symmetric Toeplitz matrix.
 * \param td     Main diagonal value of the symmetric Toeplitz matrix.
 * \param t      Remaning \a n-1 values from symmetric Toeplitz matrix.
 * \param min_k  Starting size of the Krylov subspace (starting value \c 0). Can be used to resume a previous extraction.
 * \param num_it Number of iterations to perform in the Krylov subspace.
 * \param max_k  Maximum size of the Krylov subspace. Iterations will halt if reached.
 * \param sigma  Shift value for centering eigenvalue extraction (depends on extraction interval).
 * \param p      Symmetric lanczos vectors of Krylov subspace (size align(\a n) \c x \a max_k).
 * \param q      Skew-symmetric lanczos vectors of Krylov subspace (size align(\a n) \c x \a max_k).
 * \param m      Structure array containing values from symmetric and skew-symmetric tridiagonal matrices (size \a max_k).
 * \param work   Workspace array of size >= max(5 * align(\a n), 4 * align(\a n) + \a max_k).
*/
__device__ void si2w(unsigned int n, T td, T *t, unsigned int min_k, unsigned int num_it, unsigned int max_k, T sigma,
		     T *p, T *q, tridiagonal_entry *m, T *work) {

	// Check params
	if(min_k > max_k || num_it == 0 || n == 0) return;

	// Calculate thread identifier
	volatile unsigned int tid = threadIdx.x;
	volatile unsigned int n_align = align(n);

	// Allocate required shared memory
	extern __shared__ T swork[];
	__shared__ tridiagonal_entry mk;

	// Read only beta and gamma values from previous iteration
	if(min_k > 0) {
		if(tid < 4) ((T *)&mk)[tid] = ((T *)&m[min_k - 1])[tid];
		__syncthreads();
	}

	// Initialize Lanczos vectors
	if(min_k == 0) {
		const T isq2 = (T)0.7071068;
		for(unsigned int b=0; b<n; b+=blockSize) {
			int i = b + tid;
			if(i == 0)        { p[i] = isq2; q[i] = isq2;  }
			else if(i == n-1) { p[i] = isq2; q[i] = -isq2; }
			else if(i < n)    { p[i] = (T)0.0; q[i] = (T)0.0;  }

			__syncthreads();
		}
	}

	// Assign work array (n_align-sized blocks)
	// 0: w, va
	T *w  = work;
	T *va = work;
	// 1: vs
	T *vs = work + n_align;
	// 2: v, workspace of orthogonalize (1 max_k block)
	T *v  = work + n_align * 2;
	T *work_ort = v;
	// 3: workspace of levinson (2 n_align blocks)
	T *work_lev = work + n_align * 3;

	// Main loop
	T *pk = p + min_k * n_align;
	T *qk = q + min_k * n_align;
	unsigned int max_it = min(max_k, min_k + num_it);
	__syncthreads();

	unsigned int k;
	for(k=min_k+1; k<=max_it; ++k) {

		// w = pk + qk
		for(unsigned int b=0; b<n; b+=blockSize) {
			unsigned int i = b + tid;
			if(i < n) w[i] = pk[i] + qk[i];
		}
		__syncthreads();

		// Solve (toeplitz([td t]) - s * I) * v = w
		levinson(n, td - sigma, t, w, v, work_lev, swork);

		// Calculate vs and va vectors
		axpxb_reverse_x(n,  (T)1.0, v, vs, (T)0.5, swork);
		axpxb_reverse_x(n, -(T)1.0, v, va, (T)0.5, swork);

		// Dot product: ak = vs * pk, gk = va * qk
		T aux_alpha = dot(n, vs, pk, swork);
		T aux_gamma = dot(n, va, qk, swork);

		if(tid == 0) { mk.alpha = aux_alpha; mk.gamma = aux_gamma; }
		__syncthreads();

		// Avoid unnecessary calculations if reached maximum subspace size
		if(k < max_k) {

			// vs -= ak * pk, va -= gk * qk
			axpy(n, -mk.alpha, pk, vs);
			axpy(n, -mk.gamma, qk, va);

			// vs -= b(k-1) * p(k-1), va -= d(k-1) * q(k-1)
			if(k > 1) {
				// mk.beta and mk.delta from previous iteration (not updated yet)
				axpy(n, -mk.beta,  pk - n_align, vs);
				axpy(n, -mk.delta, qk - n_align, va);
			}

			// Reorthogonalization
			reorthogonalize(n, k, p, vs, work_ort, swork);
			reorthogonalize(n, k, q, va, work_ort, swork);

			// bk = norm(vs), dk = norm(va)
			T aux_beta  = norm(n, vs, swork);
			T aux_delta = norm(n, va, swork);
			if(tid == 0) { mk.beta = aux_beta; mk.delta = aux_delta; }
			__syncthreads();

			// p(k+1) = vs / bk, q(k+1) = va / dk
			pk += n_align; qk += n_align;
			T ibk = (T)1.0 / mk.beta;
			T idk = (T)1.0 / mk.delta;

			for(unsigned int b=0; b<n; b+=blockSize) {
				volatile unsigned int i = b + tid;
				if(i < n) {
					pk[i] = vs[i] * ibk;
					qk[i] = va[i] * idk;
				}
			}
			__syncthreads();
		}

		// Store tridiagonal results in global memory
		if(tid < 4) ((T *)&m[k-1])[tid] = ((T *)&mk)[tid];
		__syncthreads();
	}
}

/**
 * Entry point for parallel shift-and-invert 2-way Lanczos evaluation.
 *
 * \param n            Size of the symmetric Toeplitz matrix.
 * \param td           Main diagonal value of the symmetric Toeplitz matrix.
 * \param t            Remaning \a n-1 values from symmetric Toeplitz matrix.
 * \param k            Current value of subspace size iterator. Real subspace size will be calculed using interval data.
 * \param inc_k        Number of iterations to perform in the Krylov subspace.
 * \param max_k        Maximum size of the Krylov subspace. Iterations will halt if reached.
 * \param exec_window  Pointer to execution window containing GPU-side descriptors of the intervals to be calculated.
*/
__global__ void si2w_parallel(unsigned int n, T td, T *t, unsigned int k, unsigned int inc_k, unsigned int max_k, interval_gpu_info *exec_window) {
	
	// Copy interval descriptor into shared memory (coalesced)
	__shared__ interval_gpu_info i;
	memcpy4(&i, exec_window + blockIdx.x, sizeof(interval_gpu_info));

	// Skip already processed intervals
	if(i.processed) return;

	// Call shift-and-invert 2-way Lanczos method
	unsigned int k_real = k - i.start_k;
	if(k_real <= max_k) si2w(n, td, t, k_real, inc_k, max_k, i.sigma, i.p, i.q, i.m, i.workspace);
}

